from .endpoints import WhisperTradesEndpoints
from .bots import WhisperTradesBots
from .variables import WhisperTradesVariables
from .scheduler import WhisperTradesScheduler
from .throttle import WhisperTradesThrottle
from .via_ui import SeleniumDriver


########################################################################################################################
########################################################################################################################
###   Selenium driver for WhisperTrades.com API                                                                      ###
###                                                                                                                  ###
###   Authored by Paul Nobrgea   Contact: Paul@PaulNobrega.net                                                       ###
###   Python Version 3.10                                                                                            ###
########################################################################################################################
########################################################################################################################
import os
import platform
import warnings
import undetected_chromedriver as uc
from selenium.webdriver.common.by import By
from selenium.webdriver.common.keys import Keys
from selenium.webdriver.support.ui import Select
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.common.exceptions import NoSuchElementException
from datetime import datetime, timedelta
import time
import copy
import chromedriver_autoinstaller


class SeleniumDriver:
    """
    Selenium-based driver for automating WhisperTrades.com UI actions.

    Handles login, bot management, and settings manipulation via browser automation.
    Designed for robust, isolated browser sessions.

    Args:
        endpts: API endpoints object for broker/bot lookups.
    """

    def _require_enabled(method):
        def wrapper(self, *args, **kwargs):
            if not getattr(self, 'is_enabled', False):
                raise Exception('Selenium Driver is not enabled. You must first call WD.via_selenium.enable(user_name, password, ...) with valid WhisperTrades credentials before using via_selenium functions.')
            return method(self, *args, **kwargs)
        return wrapper

    def __init__(self, endpts: object) -> None:
        self._endpts: object = endpts
        self.webdriver: uc.Chrome | None = None
        self.headless: bool | None = None
        self.verbose: bool | None = None
        self.user_name: str | None = None
        self.password: str | None = None
        self.delay_time_sec: int | None = None
        self._session_start_time: datetime | None = None
        self.login_url: str | None = None
        self.logout_url: str | None = None
        self.is_enabled: bool = False
        self._profile_dir: str | None = None
        return


    def __enter__(self) -> 'SeleniumDriver':
        """
        Enable use as a context manager (with statement).
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        """
        Cleanup resources when object is destroyed or context exits.
        """
        import shutil
        if self.webdriver is not None:
            self.webdriver.quit()
        if self._profile_dir:
            try:
                shutil.rmtree(self._profile_dir, ignore_errors=True)
            except Exception as e:
                warnings.warn(f'Failed to remove temp Chrome profile dir: {e}')
        return


    @_require_enabled
    def close(self) -> None:
        """
        Explicitly destroy the object and cleanup resources.
        """
        self.__exit__(None, None, None)
        return

    @_require_enabled
    def __configure_web_driver(self) -> None:
        """
        Configure and launch the Selenium browser with unique user profile for isolation.
        """

        def __get_chrome_version() -> int:
            os_name = platform.system()
            try:
                from chromedriver_autoinstaller.utils import get_chrome_version
                return int(get_chrome_version().split('.')[0])
            except Exception:
                if os_name.lower() == 'windows':
                    try:
                        return int(os.popen('reg query "HKEY_CURRENT_USER\Software\Google\Chrome\BLBeacon" /v version').read().split('REG_SZ')[-1].strip().split('.')[0])
                    except Exception:
                        raise Exception('Unable to determine Chrome version on Windows!')
                elif os_name.lower() == 'linux':
                    try:
                        ver = int(os.popen('chromium --version').read().split()[1].split('.')[0])
                    except Exception:
                        ver = int(os.popen('google-chrome --version').read().split()[-1].split('.')[0])
                    return ver
                else:
                    raise Exception('Incompatible operating system!')

        def __set_chrome_options(self_ref) -> uc.ChromeOptions:
            import tempfile
            options = uc.ChromeOptions()
            user_agent = 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/93.0.4515.159 Safari/537.36'
            options.add_argument('--no-sandbox')
            options.add_argument('--disable-dev-shm-usage')
            options.add_argument('--no-first-run')
            options.add_argument('--no-service-autorun')
            options.add_argument('--password-store=basic')
            options.add_argument('--lang=en-US')
            options.add_argument('--window-size=1920,1080')
            options.add_argument(f'--user-agent={user_agent}')
            # Add a unique user-data-dir for each instance to prevent cross talk
            self_ref._profile_dir = tempfile.mkdtemp(prefix="wt_chrome_profile_")
            options.add_argument(f'--user-data-dir={self_ref._profile_dir}')
            if self_ref.headless:
                options.add_argument('--headless')
            chrome_preferences = {'profile.default_content_settings': {"images": 2}}
            options.experimental_options["prefs"] = chrome_preferences
            return options

        try:
            chromedriver_autoinstaller.install()
        except Exception as e:
            warnings.warn(f'Could not auto-install chromedriver: {e}')

        try:
            self.webdriver = uc.Chrome(options=__set_chrome_options(self), version_main=__get_chrome_version())
        except Exception as e:
            warnings.warn(f'Could not start Chrome with detected version: {e}. Trying fallback version 114.')
            self.webdriver = uc.Chrome(options=__set_chrome_options(self), version_main=114)
        return

    @_require_enabled
    def __restart_webdriver(self) -> None:
        """
        Quit webdriver and reconfigure a new browser session.
        """
        if self.webdriver:
            self.webdriver.quit()
        self.__configure_web_driver()
        return
    
    @_require_enabled
    def __on_error_404(self, warn_except: str, error_str: str) -> bool:
        is_404 = True if 'page not found' in self.webdriver.title else False
        if is_404 and 'warn' in warn_except.lower():
            warnings.warn(error_str)
            return True
        elif is_404 and 'except' in warn_except.lower():
            raise Exception(error_str)
        return False

    @_require_enabled
    def __get_url_and_wait(self, url: str) -> bool:
        """
        Get specified url via self.webdriver. Wait for page load using Selenium's WebDriverWait.

        :param url: WhisperTrades.com url
        :type url: str
        :return: Bool value of successful url load
        :rtype: bool
        :raises Exception: If url cannot load within 2 seconds.
        """
        self.log_in_if_not()
        self.webdriver.get(url)
        try:
            WebDriverWait(self.webdriver, 2).until(lambda d: d.execute_script('return document.readyState') == 'complete')
        except Exception:
            self.webdriver.get(url)
            try:
                WebDriverWait(self.webdriver, 2).until(lambda d: d.execute_script('return document.readyState') == 'complete')
            except Exception:
                self._session_start_time = self._session_start_time - timedelta(minutes=60)
                raise Exception(f'UNABLE TO LOAD {url}')
        self._session_start_time = datetime.now()
        return False if self.__on_error_404('warn', f'Error 404: INVALID URL {url}') else True

    @_require_enabled
    def __authorization(self) -> bool:
        """
        Login to whispertrades.com via web UI login page

        :return: Bool value of authorization state
        :rtype: bool
        """

        if not self.is_enabled:
            raise Exception('Selenium Driver is not enabled.  Enabled command must be run first. Example:\nWD.via_selenium.enable(user_name: str, password: str, is_verbose: bool = False, is_headless: bool = True)')
        
        self.webdriver.get(self.login_url)
        time.sleep(1)
        username = self.webdriver.find_element(by=By.NAME, value="email")
        password = self.webdriver.find_element(by=By.NAME, value="password")
        login_btn = self.webdriver.find_element(by=By.CLASS_NAME, value="bg-green-primary")
        username.send_keys(self.user_name)
        password.send_keys(self.password)
        login_btn.click()
        time.sleep(1)
        self.__on_error_404('except', f'UNABLE TO LOGIN TO WHISPERTRADES.COM')
        self._session_start_time = datetime.now()
        return True

    def enable(self, user_name: str, password: str, is_verbose: bool = False, is_headless: bool = True) -> None:
        self.is_enabled = True
        self.webdriver = None
        self.headless = is_headless
        self.verbose = is_verbose
        self.user_name = user_name
        self.password = password
        self.delay_time_sec = 1
        self._session_start_time = datetime.now()
        self.login_url = "https://whispertrades.com/login"
        self.logout_url = "https://whispertrades.com/logout"
        self.__configure_web_driver()
        self.__authorization()
        self.current_bots = {}
        return
    

    @_require_enabled
    def is_logged_in(self) -> bool:
        """
        Determine if logged in by 1) if last session was less than 5 minutes ago or 2) presence of url redirect at login url

        :return: Bool value of logged in state
        :rtype: bool
        """
        if self._session_start_time + timedelta(minutes=2) > datetime.now():
            return True
        self.webdriver.get(self.login_url)
        time.sleep(0.3)
        return True if self.webdriver.current_url != self.login_url else False

    @_require_enabled
    def log_in_if_not(self) -> bool:
        """
        Login if not currently logged in

        :return: Bool value of logged in state
        :rtype: bool
        """
        if self.is_logged_in():
            return True
        else:
            for i in range(3):
                if self.__authorization():
                    break
            return self.is_logged_in()

    @_require_enabled
    def renew_accesss_token(self) -> bool:
        """
        Log out user and login again.

        :return: Bool value of authorization state
        :rtype: bool
        """
        self.revoke_accesss_token()
        return self.__authorization()

    @_require_enabled
    def revoke_accesss_token(self) -> None:
        """
        Log out user.
        """
        self.webdriver.get(self.logout_url)
        return
    
    @_require_enabled
    def __update_field(self, field, value) -> None:
        """
        Update a form field with a new value using keyboard shortcuts.
        Uses BACKSPACE to clear number fields, since DELETE may not work for <input type="number">.
        """
        from selenium.webdriver.common.keys import Keys
        try:
            # Try to clear with BACKSPACE (repeat 10 times)
            field.click()
            field.send_keys(Keys.END)
            for _ in range(10):
                field.send_keys(Keys.BACKSPACE)
            if value:
                field.send_keys(str(value))
        except Exception as e:
            warnings.warn(f'Failed to update field: {e}')
        return

    @_require_enabled
    def renew_schwab_connection(self, schwab_user: str, schwab_pass: str, blacklist: list=[]) -> bool:
        """
        Renew schwab connection with passed schwab credentials
        
        :param schwab_user: Schwab User Name
        :type schwab_user: string
        :param schwab_pass: Schwab Password
        :type schwab_pass: string
        :param blacklist: List of schwab WT broker numbers to not renew
        :type blacklist: list
        :return: Success as True or False
        :rtype: bool
        """

        def __find_and_wait(id='', timeout=5):
            try:
                element = WebDriverWait(self.webdriver, timeout).until(
                    EC.presence_of_element_located((By.ID, id))
                )
                if 'traceback' in element.text.lower():
                    raise NoSuchElementException
                return element
            except Exception:
                return None

        def __click_by_id_via_js(id):
            btn = __find_and_wait(id=id)
            if btn:
                return self.webdriver.execute_script ("arguments[0].click();",btn)
            return None
        
        def __click_by_id(id):
            element = __find_and_wait(id=id)
            if element:
                try:
                    return element.click()
                except Exception as e:
                    warnings.warn(f'Failed to click element by ID {id}: {e}')
            else:
                try:
                    __click_by_id_via_js(id)
                except Exception as e:
                    warnings.warn(f'Failed to click element by ID {id} via JS: {e}')
            return None

        all_brokers = self._endpts.brokers.get_all_broker_connections()
        connections = [d for d in all_brokers if d['broker'].lower() == 'schwab' and d['number'] not in blacklist]

        for c in connections:
            method = 'renew' if c['status'].lower() == 'active' else 'enable'
            schwab_renew_url = f"https://whispertrades.com/broker_connections/{c['number']}/{method}"
            print(f'Renew URL: {schwab_renew_url}')
            self.__get_url_and_wait(schwab_renew_url)
            user_field = __find_and_wait(id='loginIdInput')
            pass_field = __find_and_wait(id='passwordInput')
            login_btn = __find_and_wait(id='btnLogin')
            self.__update_field(user_field, schwab_user)
            self.__update_field(pass_field, schwab_pass)
            login_btn.click()

            new_device = None
            try:
                # Wait for either 'mobile_approve' or 'otp_sms' to appear (max 10s)
                WebDriverWait(self.webdriver, 10).until(
                    lambda d: d.find_elements(By.ID, 'mobile_approve') or d.find_elements(By.ID, 'otp_sms')
                )
                if self.webdriver.find_elements(By.ID, 'mobile_approve'):
                    id_selection = 'mobile_approve'
                    new_device = __click_by_id(id_selection)
                    try:
                        # Wait for the 'Yes, remember this device' radio button to be clickable
                        remember_radio = WebDriverWait(self.webdriver, 10).until(
                            EC.element_to_be_clickable((By.ID, 'remember-device-yes-content'))
                        )
                        remember_radio.click()
                        # Wait for the Continue button to be clickable and click it
                        continue_btn = WebDriverWait(self.webdriver, 10).until(
                            EC.element_to_be_clickable((By.ID, 'btnContinue'))
                        )
                        continue_btn.click()
                    except Exception as e:
                        warnings.warn(f'2FA remember device page not found or could not interact: {e}')
                elif self.webdriver.find_elements(By.ID, 'otp_sms'):
                    id_selection = 'otp_sms'
                    new_device = __click_by_id(id_selection)
                else:
                    warnings.warn('Neither mobile nor sms approval options found after waiting.')
                
            except Exception as e:
                warnings.warn(f'Failed to find mobile or sms approval options on page: {e}')
                new_device = None
            
            # Accept Terms Page
            if new_device:
                # If device is not yet trusted
                new_device.click()
            try:
                accept_terms = WebDriverWait(self.webdriver, 10).until(EC.element_to_be_clickable((By.ID, 'acceptTerms')))
                accept_terms.click()
                clicks = ['submit-btn', 'agree-modal-btn-', 'submit-btn']
                for c in clicks:
                    __click_by_id(c)
                confirm = WebDriverWait(self.webdriver, 10).until(EC.element_to_be_clickable((By.ID, 'submit-btn')))
                confirm.click()
            except Exception as e:
                warnings.warn(f'Failed to accept terms: {e}')
        return True

    @_require_enabled
    def get_entry_settings(self, bot_num: str) -> dict:
            """
            Get current WT Bot Entry Settings of specified bot number
            Scrapes all entry fields from the WhisperTrades bot form, including frequency, allocation, DTEs, strike selection, filters, toggles, etc.
            Handles missing/optional fields robustly.
            :param bot_num: WhisperTrades bot Number
            :type bot_num: string
            :return: Dictionary of settings
            :rtype: dict
            """

            # Helper functions must be defined before use
            def safe_find_value(by, value, attr='value', default=None):
                try:
                    el = self.webdriver.find_element(by=by, value=value)
                    return el.get_attribute(attr)
                except Exception:
                    return default

            def safe_find_checkbox(by, value, default=False):
                try:
                    el = self.webdriver.find_element(by=by, value=value)
                    return el.is_selected()
                except Exception:
                    return default

            def safe_find_select(by, value, default=None):
                try:
                    el = self.webdriver.find_element(by=by, value=value)
                    return el.get_attribute('value')
                except Exception:
                    return default

            wt_entry_edit_url = f'https://whispertrades.com/bots/{bot_num}/entry/edit'
            self.__get_url_and_wait(wt_entry_edit_url)

            # --- Allocation Section ---
            settings = {}
            # Frequency (dropdown)
            settings['frequency'] = safe_find_select(By.ID, 'form.frequency')
            # Maximum Entries Per Day (dropdown)
            settings['maximum_entries_per_day'] = safe_find_select(By.ID, 'form.maximum_entries_per_day')
            # Maximum Concurrent Positions (dropdown)
            settings['maximum_concurrent_positions'] = safe_find_select(By.ID, 'form.maximum_concurrent_positions')
            # Day of Week (dropdown)
            settings['day_of_week'] = safe_find_select(By.ID, 'form.day_of_week')
            # Allocation Type (dropdown)
            settings['allocation_type'] = safe_find_select(By.ID, 'form.allocation_type')
            # Contract Quantity (number input)
            settings['contract_quantity'] = safe_find_value(By.ID, 'form.allocation_quantity')
            # Leverage Amount
            settings['leverage_amount'] = safe_find_value(By.ID, 'form.allocation_leverage')
            # Allocation Percent 
            settings['percent_of_portfolio'] = safe_find_value(By.ID, 'form.allocation_percent')

            # --- Strike Selection Section ---
            # DTEs
            settings['minimum_days_to_expiration'] = safe_find_value(By.ID, 'form.minimum_days_to_expiration')
            settings['target_days_to_expiration'] = safe_find_value(By.ID, 'form.target_days_to_expiration')
            settings['maximum_days_to_expiration'] = safe_find_value(By.ID, 'form.maximum_days_to_expiration')

            # Put Short Strike Target Type (dropdown)
            settings['put_short_strike_type'] = safe_find_select(By.ID, 'form.put_short_strike_target_type')
            # Put Short Strike Delta fields
            settings['put_short_strike_minimum_delta'] = safe_find_value(By.ID, 'form.put_short_strike_delta_minimum')
            settings['put_short_strike_target_delta'] = safe_find_value(By.ID, 'form.put_short_strike_delta')
            settings['put_short_strike_maximum_delta'] = safe_find_value(By.ID, 'form.put_short_strike_delta_maximum')
            # Put Short Strike Premium fields
            settings['put_short_strike_minimum_premium'] = safe_find_value(By.ID, 'form.put_short_strike_premium_minimum')
            settings['put_short_strike_target_premium'] = safe_find_value(By.ID, 'form.put_short_strike_premium')
            settings['put_short_strike_maximum_premium'] = safe_find_value(By.ID, 'form.put_short_strike_premium_maximum')
            # Put Short Strike Percent OTM fields
            settings['put_short_strike_percent_otm_minimum'] = safe_find_value(By.ID, 'form.put_short_strike_percent_otm_minimum')
            settings['put_short_strike_target_percent_otm'] = safe_find_value(By.ID, 'form.put_short_strike_percent_otm')
            settings['put_short_strike_percent_otm_maximum'] = safe_find_value(By.ID, 'form.put_short_strike_percent_otm_maximum')

            # Put Spread Target Type (dropdown)
            settings['put_long_strike_type'] = safe_find_select(By.ID, 'form.put_spread_strike_target_type')
            # Put Spread Target Delta
            settings['put_long_strike_target_delta'] = safe_find_value(By.ID, 'form.put_spread_strike_target_delta')
            ### Restrict Spread Width By (dropdown)
            settings['restrict_put_spread_width_by'] = safe_find_select(By.ID, 'form.restrict_put_spread_width_by')
            ### Points
            settings['put_spread_minimum_width_points'] = safe_find_value(By.ID, 'form.put_spread_minimum_width')
            settings['put_spread_maximum_width_points'] = safe_find_value(By.ID, 'form.put_spread_maximum_width')
            ### Percent
            settings['put_spread_minimum_width_percent'] = safe_find_value(By.ID, 'form.put_spread_minimum_width_percent')
            settings['put_spread_maximum_width_percent'] = safe_find_value(By.ID, 'form.put_spread_maximum_width_percent')
            # Put Spread Target Points
            settings['put_spread_target_width_points'] = safe_find_value(By.ID, 'form.put_spread_width')
            # Put Spread Target Premium
            settings['put_spread_strike_target_premium'] = safe_find_value(By.ID, 'form.put_spread_strike_target_price')
            ### Uses same Restrict Spread Width By dropdown as above
            # Put Spread Target Percent
            settings['put_spread_target_width_percent'] = safe_find_value(By.ID, 'form.put_spread_strike_percent_from_main_strike')
            # Smart Spread Width (toggle)
            settings['put_spread_smart_width'] = safe_find_checkbox(By.ID, 'form.put_spread_smart_width')

            # Call Short Strike Target Type (dropdown)
            settings['call_short_strike_type'] = safe_find_select(By.ID, 'form.call_short_strike_target_type')
            # Call Short Strike Delta fields
            settings['call_short_strike_minimum_delta'] = safe_find_value(By.ID, 'form.call_short_strike_delta_minimum')
            settings['call_short_strike_target_delta'] = safe_find_value(By.ID, 'form.call_short_strike_delta')
            settings['call_short_strike_maximum_delta'] = safe_find_value(By.ID, 'form.call_short_strike_delta_maximum')
            # Call Short Strike Premium fields
            settings['call_short_strike_minimum_premium'] = safe_find_value(By.ID, 'form.call_short_strike_premium_minimum')
            settings['call_short_strike_target_premium'] = safe_find_value(By.ID, 'form.call_short_strike_premium')
            settings['call_short_strike_maximum_premium'] = safe_find_value(By.ID, 'form.call_short_strike_premium_maximum')
            # Call Short Strike Percent OTM fields
            settings['call_short_strike_percent_otm_minimum'] = safe_find_value(By.ID, 'form.call_short_strike_percent_otm_minimum')
            settings['call_short_strike_target_percent_otm'] = safe_find_value(By.ID, 'form.call_short_strike_percent_otm')
            settings['call_short_strike_percent_otm_maximum'] = safe_find_value(By.ID, 'form.call_short_strike_percent_otm_maximum')

            # Call Spread Target Type (dropdown)
            settings['call_long_strike_type'] = safe_find_select(By.ID, 'form.call_spread_strike_target_type')
            # Call Spread Target Delta
            settings['call_long_strike_target_delta'] = safe_find_value(By.ID, 'form.call_spread_strike_target_delta')
            ### Restrict Spread Width By (dropdown)
            settings['restrict_call_spread_width_by'] = safe_find_select(By.ID, 'form.restrict_call_spread_width_by')
            ### Points
            settings['call_spread_minimum_width_points'] = safe_find_value(By.ID, 'form.call_spread_minimum_width')
            settings['call_spread_maximum_width_points'] = safe_find_value(By.ID, 'form.call_spread_maximum_width')
            ### Percent
            settings['call_spread_minimum_width_percent'] = safe_find_value(By.ID, 'form.call_spread_minimum_width_percent')
            settings['call_spread_maximum_width_percent'] = safe_find_value(By.ID, 'form.call_spread_maximum_width_percent')
            # Call Spread Target Points
            settings['call_spread_target_width_points'] = safe_find_value(By.ID, 'form.call_spread_width')
            # Call Spread Target Premium
            settings['call_spread_strike_target_premium'] = safe_find_value(By.ID, 'form.call_spread_strike_target_price')
            ### Uses same Restrict Spread Width By dropdown as above
            # Call Spread Target Percent
            settings['call_spread_target_width_percent'] = safe_find_value(By.ID, 'form.call_spread_strike_percent_from_main_strike')
            # Smart Spread Width (toggle)
            settings['call_spread_smart_width'] = safe_find_checkbox(By.ID, 'form.call_spread_smart_width')

            # Entry Filters (expanded to all possible fields from examples.txt)
            settings['minimum_starting_premium'] = safe_find_value(By.ID, 'form.minimum_premium')
            settings['maximum_starting_premium'] = safe_find_value(By.ID, 'form.maximum_premium')
            settings['minimum_iv'] = safe_find_value(By.ID, 'form.minimum_iv')
            settings['maximum_iv'] = safe_find_value(By.ID, 'form.maximum_iv')
            settings['minimum_vix'] = safe_find_value(By.ID, 'form.minimum_vix')
            settings['maximum_vix'] = safe_find_value(By.ID, 'form.maximum_vix')
            settings['minimum_underlying_percent_move_from_close'] = safe_find_value(By.ID, 'form.minimum_underlying_percent_move_from_close')
            settings['maximum_underlying_percent_move_from_close'] = safe_find_value(By.ID, 'form.maximum_underlying_percent_move_from_close')
            settings['minimum_underlying_percent_move_from_open'] = safe_find_value(By.ID, 'form.minimum_underlying_percent_move_from_open')
            settings['maximum_underlying_percent_move_from_open'] = safe_find_value(By.ID, 'form.maximum_underlying_percent_move_from_open')
            settings['entry_percent_from_today_high_toggle'] = safe_find_checkbox(By.ID, 'form.entry_percent_from_today_high_toggle')
            settings['minimum_underlying_percent_move_from_today_high'] = safe_find_value(By.ID, 'form.entry_minimum_percent_move_from_today_high')
            settings['maximum_underlying_percent_move_from_today_high'] = safe_find_value(By.ID, 'form.entry_maximum_percent_move_from_today_high')
            settings['underlying_percent_from_today_high_start_time'] = safe_find_select(By.ID, 'form.entry_percent_from_today_high_start_time')
            settings['underlying_percent_from_today_high_end_time'] = safe_find_select(By.ID, 'form.entry_percent_from_today_high_end_time')
            settings['entry_percent_from_today_low_toggle'] = safe_find_checkbox(By.ID, 'form.entry_percent_from_today_low_toggle')
            settings['minimum_underlying_percent_move_from_today_low'] = safe_find_value(By.ID, 'form.entry_minimum_percent_move_from_today_low')
            settings['maximum_underlying_percent_move_from_today_low'] = safe_find_value(By.ID, 'form.entry_maximum_percent_move_from_today_low')
            settings['underlying_percent_from_today_low_start_time'] = safe_find_select(By.ID, 'form.entry_percent_from_today_low_start_time')
            settings['underlying_percent_from_today_low_end_time'] = safe_find_select(By.ID, 'form.entry_percent_from_today_low_end_time')
            settings['entry_percent_from_prior_high_toggle'] = safe_find_checkbox(By.ID, 'form.entry_percent_from_prior_high_toggle')
            settings['minimum_underlying_percent_move_from_prior_high'] = safe_find_value(By.ID, 'form.entry_minimum_percent_move_from_prior_high')
            settings['maximum_underlying_percent_move_from_prior_high'] = safe_find_value(By.ID, 'form.entry_maximum_percent_move_from_prior_high')
            settings['underlying_percent_from_prior_high_start_time'] = safe_find_select(By.ID, 'form.entry_percent_from_prior_high_start_time')
            settings['underlying_percent_from_prior_high_end_time'] = safe_find_select(By.ID, 'form.entry_percent_from_prior_high_end_time')
            settings['entry_percent_from_prior_low_toggle'] = safe_find_checkbox(By.ID, 'form.entry_percent_from_prior_low_toggle')
            settings['minimum_underlying_percent_move_from_prior_low'] = safe_find_value(By.ID, 'form.entry_minimum_percent_move_from_prior_low')
            settings['maximum_underlying_percent_move_from_prior_low'] = safe_find_value(By.ID, 'form.entry_maximum_percent_move_from_prior_low')
            settings['underlying_percent_from_prior_low_start_time'] = safe_find_select(By.ID, 'form.entry_percent_from_prior_low_start_time')
            settings['underlying_percent_from_prior_low_end_time'] = safe_find_select(By.ID, 'form.entry_percent_from_prior_low_end_time')
            settings['entry_ma_crossover_toggle'] = safe_find_checkbox(By.ID, 'form.entry_ma_crossover_toggle')
            settings['ma_crossover_one'] = safe_find_select(By.ID, 'form.entry_ma_crossover_one')
            settings['ma_crossover_percent'] = safe_find_value(By.ID, 'form.entry_ma_crossover_percent')
            settings['ma_crossover_type'] = safe_find_select(By.ID, 'form.entry_ma_crossover_type')
            settings['ma_crossover_two'] = safe_find_select(By.ID, 'form.entry_ma_crossover_two')
            settings['entry_ma_value_toggle'] = safe_find_checkbox(By.ID, 'form.entry_ma_value_toggle')
            settings['ma_value_percent'] = safe_find_value(By.ID, 'form.entry_ma_value_percent')
            settings['ma_value_type'] = safe_find_select(By.ID, 'form.entry_ma_value_type')
            settings['ma_value_ma'] = safe_find_select(By.ID, 'form.entry_ma_value_ma')
            settings['avoid_fomc'] = safe_find_checkbox(By.ID, 'form.avoid_fomc')
            settings['avoid_fomc_days_before'] = safe_find_value(By.ID, 'form.avoid_fomc_days_before')
            settings['avoid_fomc_days_after'] = safe_find_value(By.ID, 'form.avoid_fomc_days_after')
            settings['min_underlying_price'] = safe_find_value(By.ID, 'form.minimum_underlying_price')
            settings['max_underlying_price'] = safe_find_value(By.ID, 'form.maximum_underlying_price')
            settings['min_underlying_iv'] = safe_find_value(By.ID, 'form.minimum_underlying_iv')
            settings['max_underlying_iv'] = safe_find_value(By.ID, 'form.maximum_underlying_iv')
            settings['min_underlying_iv_rank'] = safe_find_value(By.ID, 'form.minimum_underlying_iv_rank')
            settings['max_underlying_iv_rank'] = safe_find_value(By.ID, 'form.maximum_underlying_iv_rank')
            settings['min_underlying_iv_percentile'] = safe_find_value(By.ID, 'form.minimum_underlying_iv_percentile')
            settings['max_underlying_iv_percentile'] = safe_find_value(By.ID, 'form.maximum_underlying_iv_percentile')
            settings['skip_earnings'] = safe_find_checkbox(By.ID, 'form.skip_earnings')
            settings['exclude_tickers'] = safe_find_value(By.ID, 'form.exclude_tickers')
            settings['only_tickers'] = safe_find_value(By.ID, 'form.only_tickers')
            settings['min_underlying_price_change'] = safe_find_value(By.ID, 'form.minimum_underlying_price_change')
            settings['max_underlying_price_change'] = safe_find_value(By.ID, 'form.maximum_underlying_price_change')
            settings['min_underlying_volume'] = safe_find_value(By.ID, 'form.minimum_underlying_volume')
            settings['min_underlying_open_interest'] = safe_find_value(By.ID, 'form.minimum_underlying_open_interest')
            settings['min_underlying_market_cap'] = safe_find_value(By.ID, 'form.minimum_underlying_market_cap')
            settings['underlying_sector'] = safe_find_select(By.ID, 'form.underlying_sector')
            settings['underlying_industry'] = safe_find_select(By.ID, 'form.underlying_industry')
            settings['only_etfs'] = safe_find_checkbox(By.ID, 'form.only_etfs')
            settings['only_stocks'] = safe_find_checkbox(By.ID, 'form.only_stocks')
            settings['only_index'] = safe_find_checkbox(By.ID, 'form.only_index')
            settings['only_liquid_options'] = safe_find_checkbox(By.ID, 'form.only_liquid_options')
            settings['only_marginable'] = safe_find_checkbox(By.ID, 'form.only_marginable')
            settings['only_shortable'] = safe_find_checkbox(By.ID, 'form.only_shortable')
            settings['only_easy_to_borrow'] = safe_find_checkbox(By.ID, 'form.only_easy_to_borrow')
            settings['only_hard_to_borrow'] = safe_find_checkbox(By.ID, 'form.only_hard_to_borrow')
            settings['custom_filter'] = safe_find_value(By.ID, 'form.custom_filter')

            # --- Miscellaneous Section ---
            # Entry Speed (dropdown)
            settings['entry_speed'] = safe_find_select(By.ID, 'form.entry_speed')
            # Move Strike Selection With Conflict (toggle)
            settings['move_strike_selection_with_conflict'] = safe_find_checkbox(By.ID, 'form.move_strike_selection_with_conflict')

            # Scrape all entryVariables repeater fields (bot_variable_id, condition, value)
            entry_variables = []
            try:
                idx = 0
                while True:
                    prefix = f'form.entryVariables.{idx}.'
                    bot_variable_id = safe_find_select(By.ID, prefix + 'bot_variable_id')
                    condition = safe_find_select(By.ID, prefix + 'condition')
                    value = safe_find_value(By.ID, prefix + 'value')
                    if bot_variable_id is None and condition is None and value is None:
                        break
                    entry_variables.append({
                        'bot_variable_id': bot_variable_id,
                        'condition': condition,
                        'value': value
                    })
                    idx += 1
            except Exception:
                pass
            settings['variables'] = entry_variables

            # Entry time window (dropdowns) -- updated to match actual HTML
            try:
                earliest_entry_el = self.webdriver.find_element(By.ID, 'form.earliest_entry_time')
                latest_entry_el = self.webdriver.find_element(By.ID, 'form.latest_entry_time')
                earliest_entry = Select(earliest_entry_el).first_selected_option.text.strip()
                latest_entry = Select(latest_entry_el).first_selected_option.text.strip()
                settings['earliest_entry_time'] = earliest_entry if earliest_entry else None
                settings['latest_entry_time'] = latest_entry if latest_entry else None
            except Exception:
                settings['earliest_entry_time'] = None
                settings['latest_entry_time'] = None

            # Entry days of week (checkboxes) -- updated to match actual HTML
            days = ['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday']
            checked_days = []
            try:
                els = self.webdriver.find_elements(By.CSS_SELECTOR, "input[type='checkbox'][wire\\:model='data.days_of_week']")
                if not els:
                    print('[DEBUG] No checkboxes found for days_of_week')
                for el in els:
                    val = el.get_attribute('value')
                    if val in days and el.is_selected():
                        checked_days.append(val)
            except Exception as e:
                print(f'[DEBUG] Exception in days_of_week scraping: {e}')
            # Always return a list, never None
            settings['days_of_week'] = checked_days



            # --- Build API-compatible output dict ---
            def yesno(val):
                if val is True:
                    return 'Yes'
                if val is False:
                    return 'No'
                return None

            # Convert days_of_week to API format
            api_days_of_week = None
            if 'days_of_week' in settings:
                if isinstance(settings['days_of_week'], list):
                    if len(settings['days_of_week']) == 5:
                        api_days_of_week = 'All'
                    elif len(settings['days_of_week']) > 0:
                        api_days_of_week = settings['days_of_week']
                        
            # Build API-compatible output dict with all available settings
            api_settings = {}
            for k, v in settings.items():
                # Convert booleans to 'Yes'/'No' for API where appropriate
                if isinstance(v, bool):
                    api_settings[k] = yesno(v)
                else:
                    api_settings[k] = v
            # Preserve days_of_week API format
            api_settings['days_of_week'] = api_days_of_week

            return api_settings


    @_require_enabled
    def update_entry_settings(self, bot_num: str, entry_settings_dict: dict) -> dict:
        """
        Update WT Bot Entry Settings for the specified bot number using a dictionary of values.
        Updates all entry fields present in the UI, including checkboxes, selects, text fields, and repeaters.
        :param bot_num: WhisperTrades bot Number
        :param entry_settings_dict: Dictionary with the same format as returned by get_entry_settings
        :return: Dictionary of changed settings
        :rtype: dict
        """
        # Get current settings and merge with new values
        settings = self.get_entry_settings(bot_num)
        initial_settings = copy.deepcopy(settings)
        settings.update(entry_settings_dict)

        if initial_settings == settings:
            if self.verbose:
                print(f'No changes required for ENTRY settings of bot: {bot_num}')
            return settings

        wt_entry_edit_url = f'https://whispertrades.com/bots/{bot_num}/entry/edit'
        self.__get_url_and_wait(wt_entry_edit_url)
        save_btn = self.webdriver.find_element(by=By.CLASS_NAME, value='bg-green-600')

        from selenium.webdriver.common.keys import Keys
        def safe_update_field(field_id, value, input_type='text'):
            try:
                el = self.webdriver.find_element(By.ID, field_id)
                if input_type == 'checkbox':
                    if bool(el.is_selected()) != bool(value):
                        el.click()
                elif input_type == 'select':
                    sel = Select(el)
                    try:
                        sel.select_by_value(str(value))
                    except Exception:
                        for option in sel.options:
                            if option.text.strip() == str(value).strip():
                                option.click()
                                break
                else:
                    # Use the known working __update_field for text fields, which now uses BACKSPACE
                    if value is None:
                        self.__update_field(el, "")
                    else:
                        self.__update_field(el, str(value))
                time.sleep(0.1)
            except Exception:
                pass

        # Map of dependent fields: {dependent_key: controller_key}
        dependent_fields = {
            'allocation_percent': 'allocation_type',
            'leverage_amount': 'allocation_type',
            'put_spread_minimum_width_percent': 'restrict_put_spread_width_by',
            'put_spread_maximum_width_percent': 'restrict_put_spread_width_by',
            'put_spread_minimum_width_points': 'restrict_put_spread_width_by',
            'put_spread_maximum_width_points': 'restrict_put_spread_width_by',
            'put_spread_smart_width': 'restrict_put_spread_width_by',
            'call_spread_minimum_width_percent': 'restrict_call_spread_width_by',
            'call_spread_maximum_width_percent': 'restrict_call_spread_width_by',
            'call_spread_minimum_width_points': 'restrict_call_spread_width_by',
            'call_spread_maximum_width_points': 'restrict_call_spread_width_by',
            'call_spread_smart_width': 'restrict_call_spread_width_by',
        }

        # Field map: key -> (element_id, type)
        field_map = {
            'frequency': ('form.frequency', 'select'),
            'maximum_entries_per_day': ('form.maximum_entries_per_day', 'select'),
            'maximum_concurrent_positions': ('form.maximum_concurrent_positions', 'select'),
            'day_of_week': ('form.day_of_week', 'select'),
            'allocation_type': ('form.allocation_type', 'select'),
            'contract_quantity': ('form.allocation_quantity', 'text'),
            'leverage_amount': ('form.allocation_leverage', 'text'),
            'percent_of_portfolio': ('form.allocation_percent', 'text'),
            'minimum_days_to_expiration': ('form.minimum_days_to_expiration', 'text'),
            'target_days_to_expiration': ('form.target_days_to_expiration', 'text'),
            'maximum_days_to_expiration': ('form.maximum_days_to_expiration', 'text'),
            'put_short_strike_type': ('form.put_short_strike_target_type', 'select'),
            'put_short_strike_minimum_delta': ('form.put_short_strike_delta_minimum', 'text'),
            'put_short_strike_target_delta': ('form.put_short_strike_delta', 'text'),
            'put_short_strike_maximum_delta': ('form.put_short_strike_delta_maximum', 'text'),
            'put_short_strike_minimum_premium': ('form.put_short_strike_premium_minimum', 'text'),
            'put_short_strike_target_premium': ('form.put_short_strike_premium', 'text'),
            'put_short_strike_maximum_premium': ('form.put_short_strike_premium_maximum', 'text'),
            'put_short_strike_percent_otm_minimum': ('form.put_short_strike_percent_otm_minimum', 'text'),
            'put_short_strike_target_percent_otm': ('form.put_short_strike_percent_otm', 'text'),
            'put_short_strike_percent_otm_maximum': ('form.put_short_strike_percent_otm_maximum', 'text'),
            'put_long_strike_type': ('form.put_spread_strike_target_type', 'select'),
            'put_long_strike_target_delta': ('form.put_spread_strike_target_delta', 'text'),
            'restrict_put_spread_width_by': ('form.restrict_put_spread_width_by', 'select'),
            'put_spread_minimum_width_points': ('form.put_spread_minimum_width', 'text'),
            'put_spread_maximum_width_points': ('form.put_spread_maximum_width', 'text'),
            'put_spread_minimum_width_percent': ('form.put_spread_minimum_width_percent', 'text'),
            'put_spread_maximum_width_percent': ('form.put_spread_maximum_width_percent', 'text'),
            'put_spread_target_width_points': ('form.put_spread_width', 'text'),
            'put_spread_strike_target_premium': ('form.put_spread_strike_target_price', 'text'),
            'put_spread_target_width_percent': ('form.put_spread_strike_percent_from_main_strike', 'text'),
            'put_spread_smart_width': ('form.put_spread_smart_width', 'checkbox'),
            'call_short_strike_type': ('form.call_short_strike_target_type', 'select'),
            'call_short_strike_minimum_delta': ('form.call_short_strike_delta_minimum', 'text'),
            'call_short_strike_target_delta': ('form.call_short_strike_delta', 'text'),
            'call_short_strike_maximum_delta': ('form.call_short_strike_delta_maximum', 'text'),
            'call_short_strike_minimum_premium': ('form.call_short_strike_premium_minimum', 'text'),
            'call_short_strike_target_premium': ('form.call_short_strike_premium', 'text'),
            'call_short_strike_maximum_premium': ('form.call_short_strike_premium_maximum', 'text'),
            'call_short_strike_percent_otm_minimum': ('form.call_short_strike_percent_otm_minimum', 'text'),
            'call_short_strike_target_percent_otm': ('form.call_short_strike_percent_otm', 'text'),
            'call_short_strike_percent_otm_maximum': ('form.call_short_strike_percent_otm_maximum', 'text'),
            'call_long_strike_type': ('form.call_spread_strike_target_type', 'select'),
            'call_long_strike_target_delta': ('form.call_spread_strike_target_delta', 'text'),
            'restrict_call_spread_width_by': ('form.restrict_call_spread_width_by', 'select'),
            'call_spread_minimum_width_points': ('form.call_spread_minimum_width', 'text'),
            'call_spread_maximum_width_points': ('form.call_spread_maximum_width', 'text'),
            'call_spread_minimum_width_percent': ('form.call_spread_minimum_width_percent', 'text'),
            'call_spread_maximum_width_percent': ('form.call_spread_maximum_width_percent', 'text'),
            'call_spread_target_width_points': ('form.call_spread_width', 'text'),
            'call_spread_strike_target_premium': ('form.call_spread_strike_target_price', 'text'),
            'call_spread_target_width_percent': ('form.call_spread_strike_percent_from_main_strike', 'text'),
            'call_spread_smart_width': ('form.call_spread_smart_width', 'checkbox'),
            # Entry Filters and Miscellaneous
            'minimum_starting_premium': ('form.minimum_premium', 'text'),
            'maximum_starting_premium': ('form.maximum_premium', 'text'),
            'minimum_iv': ('form.minimum_iv', 'text'),
            'maximum_iv': ('form.maximum_iv', 'text'),
            'minimum_vix': ('form.minimum_vix', 'text'),
            'maximum_vix': ('form.maximum_vix', 'text'),
            'minimum_underlying_percent_move_from_close': ('form.minimum_underlying_percent_move_from_close', 'text'),
            'maximum_underlying_percent_move_from_close': ('form.maximum_underlying_percent_move_from_close', 'text'),
            'minimum_underlying_percent_move_from_open': ('form.minimum_underlying_percent_move_from_open', 'text'),
            'maximum_underlying_percent_move_from_open': ('form.maximum_underlying_percent_move_from_open', 'text'),
            'entry_percent_from_today_high_toggle': ('form.entry_percent_from_today_high_toggle', 'checkbox'),
            'minimum_underlying_percent_move_from_today_high': ('form.entry_minimum_percent_move_from_today_high', 'text'),
            'maximum_underlying_percent_move_from_today_high': ('form.entry_maximum_percent_move_from_today_high', 'text'),
            'underlying_percent_from_today_high_start_time': ('form.entry_percent_from_today_high_start_time', 'select'),
            'underlying_percent_from_today_high_end_time': ('form.entry_percent_from_today_high_end_time', 'select'),
            'entry_percent_from_today_low_toggle': ('form.entry_percent_from_today_low_toggle', 'checkbox'),
            'minimum_underlying_percent_move_from_today_low': ('form.entry_minimum_percent_move_from_today_low', 'text'),
            'maximum_underlying_percent_move_from_today_low': ('form.entry_maximum_percent_move_from_today_low', 'text'),
            'underlying_percent_from_today_low_start_time': ('form.entry_percent_from_today_low_start_time', 'select'),
            'underlying_percent_from_today_low_end_time': ('form.entry_percent_from_today_low_end_time', 'select'),
            'entry_percent_from_prior_high_toggle': ('form.entry_percent_from_prior_high_toggle', 'checkbox'),
            'minimum_underlying_percent_move_from_prior_high': ('form.entry_minimum_percent_move_from_prior_high', 'text'),
            'maximum_underlying_percent_move_from_prior_high': ('form.entry_maximum_percent_move_from_prior_high', 'text'),
            'underlying_percent_from_prior_high_start_time': ('form.entry_percent_from_prior_high_start_time', 'select'),
            'underlying_percent_from_prior_high_end_time': ('form.entry_percent_from_prior_high_end_time', 'select'),
            'entry_percent_from_prior_low_toggle': ('form.entry_percent_from_prior_low_toggle', 'checkbox'),
            'minimum_underlying_percent_move_from_prior_low': ('form.entry_minimum_percent_move_from_prior_low', 'text'),
            'maximum_underlying_percent_move_from_prior_low': ('form.entry_maximum_percent_move_from_prior_low', 'text'),
            'underlying_percent_from_prior_low_start_time': ('form.entry_percent_from_prior_low_start_time', 'select'),
            'underlying_percent_from_prior_low_end_time': ('form.entry_percent_from_prior_low_end_time', 'select'),
            'entry_ma_crossover_toggle': ('form.entry_ma_crossover_toggle', 'checkbox'),
            'ma_crossover_one': ('form.entry_ma_crossover_one', 'select'),
            'ma_crossover_percent': ('form.entry_ma_crossover_percent', 'text'),
            'ma_crossover_type': ('form.entry_ma_crossover_type', 'select'),
            'ma_crossover_two': ('form.entry_ma_crossover_two', 'select'),
            'entry_ma_value_toggle': ('form.entry_ma_value_toggle', 'checkbox'),
            'ma_value_percent': ('form.entry_ma_value_percent', 'text'),
            'ma_value_type': ('form.entry_ma_value_type', 'select'),
            'ma_value_ma': ('form.entry_ma_value_ma', 'select'),
            'avoid_fomc': ('form.avoid_fomc', 'checkbox'),
            'avoid_fomc_days_before': ('form.avoid_fomc_days_before', 'text'),
            'avoid_fomc_days_after': ('form.avoid_fomc_days_after', 'text'),
            'min_underlying_price': ('form.minimum_underlying_price', 'text'),
            'max_underlying_price': ('form.maximum_underlying_price', 'text'),
            'min_underlying_iv': ('form.minimum_underlying_iv', 'text'),
            'max_underlying_iv': ('form.maximum_underlying_iv', 'text'),
            'min_underlying_iv_rank': ('form.minimum_underlying_iv_rank', 'text'),
            'max_underlying_iv_rank': ('form.maximum_underlying_iv_rank', 'text'),
            'min_underlying_iv_percentile': ('form.minimum_underlying_iv_percentile', 'text'),
            'max_underlying_iv_percentile': ('form.maximum_underlying_iv_percentile', 'text'),
            'skip_earnings': ('form.skip_earnings', 'checkbox'),
            'exclude_tickers': ('form.exclude_tickers', 'text'),
            'only_tickers': ('form.only_tickers', 'text'),
            'min_underlying_price_change': ('form.minimum_underlying_price_change', 'text'),
            'max_underlying_price_change': ('form.maximum_underlying_price_change', 'text'),
            'min_underlying_volume': ('form.minimum_underlying_volume', 'text'),
            'min_underlying_open_interest': ('form.minimum_underlying_open_interest', 'text'),
            'min_underlying_market_cap': ('form.minimum_underlying_market_cap', 'text'),
            'underlying_sector': ('form.underlying_sector', 'select'),
            'underlying_industry': ('form.underlying_industry', 'select'),
            'only_etfs': ('form.only_etfs', 'checkbox'),
            'only_stocks': ('form.only_stocks', 'checkbox'),
            'only_index': ('form.only_index', 'checkbox'),
            'only_liquid_options': ('form.only_liquid_options', 'checkbox'),
            'only_marginable': ('form.only_marginable', 'checkbox'),
            'only_shortable': ('form.only_shortable', 'checkbox'),
            'only_easy_to_borrow': ('form.only_easy_to_borrow', 'checkbox'),
            'only_hard_to_borrow': ('form.only_hard_to_borrow', 'checkbox'),
            'custom_filter': ('form.custom_filter', 'text'),
            # Miscellaneous
            'entry_speed': ('form.entry_speed', 'select'),
            'move_strike_selection_with_conflict': ('form.move_strike_selection_with_conflict', 'checkbox'),
        }

        # First, set all controller fields that are present in the input
        controllers_set = set()
        for dep, ctrl in dependent_fields.items():
            if ctrl in entry_settings_dict and ctrl not in controllers_set:
                val = entry_settings_dict[ctrl]
                if ctrl in field_map and val is not None:
                    safe_update_field(field_map[ctrl][0], val, field_map[ctrl][1])
                    controllers_set.add(ctrl)
                    time.sleep(0.2)  # allow UI to update

        # Now update all other fields present in the input dict
        # Map toggles to their dependent fields
        toggle_dependents = {
            'entry_percent_from_today_high_toggle': [
                'minimum_underlying_percent_move_from_today_high',
                'maximum_underlying_percent_move_from_today_high',
                'underlying_percent_from_today_high_start_time',
                'underlying_percent_from_today_high_end_time',
            ],
            'entry_percent_from_today_low_toggle': [
                'minimum_underlying_percent_move_from_today_low',
                'maximum_underlying_percent_move_from_today_low',
                'underlying_percent_from_today_low_start_time',
                'underlying_percent_from_today_low_end_time',
            ],
            'entry_percent_from_prior_high_toggle': [
                'minimum_underlying_percent_move_from_prior_high',
                'maximum_underlying_percent_move_from_prior_high',
                'underlying_percent_from_prior_high_start_time',
                'underlying_percent_from_prior_high_end_time',
            ],
            'entry_percent_from_prior_low_toggle': [
                'minimum_underlying_percent_move_from_prior_low',
                'maximum_underlying_percent_move_from_prior_low',
                'underlying_percent_from_prior_low_start_time',
                'underlying_percent_from_prior_low_end_time',
            ],
            'entry_ma_crossover_toggle': [
                'ma_crossover_one', 'ma_crossover_percent', 'ma_crossover_type', 'ma_crossover_two'
            ],
            'entry_ma_value_toggle': [
                'ma_value_percent', 'ma_value_type', 'ma_value_ma'
            ],
        }

        # First, check toggles: if enabled but all dependents are blank/missing, turn off toggle
        for toggle, dependents in toggle_dependents.items():
            if toggle in entry_settings_dict and entry_settings_dict[toggle]:
                all_blank = True
                for dep in dependents:
                    val = entry_settings_dict.get(dep)
                    if val not in (None, ''):
                        all_blank = False
                        break
                if all_blank:
                    # Turn off toggle in UI and dict
                    safe_update_field(field_map[toggle][0], False, field_map[toggle][1])
                    entry_settings_dict[toggle] = False

        # Now update all toggles and immediately set their dependents if toggle is True
        for toggle, dependents in toggle_dependents.items():
            if toggle in entry_settings_dict and entry_settings_dict[toggle]:
                # Set the toggle
                safe_update_field(field_map[toggle][0], True, field_map[toggle][1])
                # For each dependent, set from input or use a safe default
                for dep in dependents:
                    val = entry_settings_dict.get(dep)
                    if val is None or val == '':
                        # Use a safe default: '' for text, first option for select
                        field_type = field_map[dep][1]
                        if field_type == 'select':
                            try:
                                el = self.webdriver.find_element(By.ID, field_map[dep][0])
                                sel = Select(el)
                                if sel.options:
                                    sel.options[0].click()
                            except Exception:
                                pass
                        else:
                            safe_update_field(field_map[dep][0], '', field_type)
                    else:
                        safe_update_field(field_map[dep][0], val, field_map[dep][1])

        # Now update all other fields present in the input dict (skip those handled above)
        toggled_fields = set()
        for dependents in toggle_dependents.values():
            toggled_fields.update(dependents)
        toggled_fields.update(toggle_dependents.keys())
        for k, v in entry_settings_dict.items():
            if k in field_map and k not in toggled_fields:
                # If this is a dependent field, ensure its controller is set first
                if k in dependent_fields:
                    ctrl = dependent_fields[k]
                    if ctrl in entry_settings_dict and ctrl not in controllers_set:
                        safe_update_field(field_map[ctrl][0], entry_settings_dict[ctrl], field_map[ctrl][1])
                        controllers_set.add(ctrl)
                        time.sleep(0.2)
                safe_update_field(field_map[k][0], v, field_map[k][1])

        # Handle days_of_week checkboxes if present
        if 'days_of_week' in entry_settings_dict and entry_settings_dict['days_of_week'] is not None:
            try:
                days = entry_settings_dict['days_of_week']
                els = self.webdriver.find_elements(By.CSS_SELECTOR, "input[type='checkbox'][wire\\:model='data.days_of_week']")
                for el in els:
                    val = el.get_attribute('value')
                    if val in days and not el.is_selected():
                        el.click()
                    elif val not in days and el.is_selected():
                        el.click()
            except Exception as e:
                if self.verbose:
                    print(f"[DEBUG] Exception updating days_of_week: {e}")

        # Handle earliest/latest entry time dropdowns if present
        for time_field, dom_id in [('earliest_entry_time', 'form.earliest_entry_time'), ('latest_entry_time', 'form.latest_entry_time')]:
            if time_field in entry_settings_dict and entry_settings_dict[time_field] is not None:
                try:
                    el = self.webdriver.find_element(By.ID, dom_id)
                    sel = Select(el)
                    for option in sel.options:
                        if option.text.strip() == str(entry_settings_dict[time_field]).strip():
                            option.click()
                            break
                except Exception as e:
                    if self.verbose:
                        print(f"[DEBUG] Exception updating {time_field}: {e}")

        # Handle entry variables (repeater)
        if 'variables' in entry_settings_dict and isinstance(entry_settings_dict['variables'], list):
            for idx, var in enumerate(entry_settings_dict['variables']):
                prefix = f'form.entryVariables.{idx}.'
                if 'bot_variable_id' in var and var['bot_variable_id'] is not None:
                    try:
                        el = self.webdriver.find_element(By.ID, prefix + 'bot_variable_id')
                        sel = Select(el)
                        sel.select_by_value(str(var['bot_variable_id']))
                    except Exception:
                        pass
                if 'condition' in var and var['condition'] is not None:
                    try:
                        el = self.webdriver.find_element(By.ID, prefix + 'condition')
                        sel = Select(el)
                        sel.select_by_value(str(var['condition']))
                    except Exception:
                        pass
                if 'value' in var:
                    try:
                        el = self.webdriver.find_element(By.ID, prefix + 'value')
                        if var['value'] is None:
                            self.__update_field(el, "")
                        else:
                            self.__update_field(el, str(var['value']))
                    except Exception:
                        pass

        save_btn.click()
        time.sleep(1)
        new_settings = self.get_entry_settings(bot_num)
        if self.verbose:
            print(f'Updated ENTRY settings for bot: {bot_num}')
        return new_settings

    @_require_enabled
    def get_exit_settings(self, bot_num: str) -> dict:
        """
        Get current WT Bot Exit Settings of specified bot number
        Scrapes all exit fields from the WhisperTrades bot exit form, including all standing profit targets, monitored stops, trailing stops, sensitivities, toggles, market conditions, and advanced variables.
        Handles missing/optional fields robustly.
        :param bot_num: WhisperTrades bot Number
        :type bot_num: string
        :return: Dictionary of settings
        :rtype: dict
        """
        wt_exit_edit_url = f'https://whispertrades.com/bots/{bot_num}/exit/edit'
        self.__get_url_and_wait(wt_exit_edit_url)

        def safe_find_value(by, value, attr='value', default=None):
            try:
                el = self.webdriver.find_element(by=by, value=value)
                return el.get_attribute(attr) if el.get_attribute(attr) != '' else default
            except Exception:
                return default

        def safe_find_checkbox(by, value, default=False):
            try:
                el = self.webdriver.find_element(by=by, value=value)
                return el.is_selected()
            except Exception:
                return default

        def safe_find_select(by, value, default=None):
            try:
                el = self.webdriver.find_element(by=by, value=value)
                return el.get_attribute('value')
            except Exception:
                return default

        settings = {}
        # --- Profit (Standing) Section ---
        settings['profit_target_percent'] = safe_find_value(By.ID, 'form.profit_target_percent')
        settings['profit_premium_value'] = safe_find_value(By.ID, 'form.premium_value_profit')

        # --- Monitored Stops Section ---
        settings['stop_loss_percent'] = safe_find_value(By.ID, 'form.stop_loss_percent')
        settings['loss_premium_value'] = safe_find_value(By.ID, 'form.premium_value_loss')
        settings['itm_percent_stop'] = safe_find_value(By.ID, 'form.itm_percent')
        settings['otm_percent_stop'] = safe_find_value(By.ID, 'form.otm_percent')
        settings['delta_stop'] = safe_find_value(By.ID, 'form.delta_value')
        settings['monitored_stop_sensitivity'] = safe_find_select(By.ID, 'form.monitored_stop_sensitivity')

        # --- Exit Variables (Advanced) repeater ---
        exit_vars = []
        idx = 0
        while True:
            var_prefix = f'form.exitVariables.{idx}.'
            bot_variable_id = safe_find_select(By.ID, var_prefix + 'bot_variable_id')
            condition = safe_find_select(By.ID, var_prefix + 'condition')
            value = safe_find_value(By.ID, var_prefix + 'value')
            if bot_variable_id is None and condition is None and value is None:
                break
            exit_vars.append({
                'bot_variable_id': bot_variable_id,
                'condition': condition,
                'value': value
            })
            idx += 1
        settings['variables'] = exit_vars

        # --- Trailing Stops Section ---
        settings['trail_profit_percent_trigger'] = safe_find_value(By.ID, 'form.trail_profit_target_percent_trigger')
        settings['trail_profit_percent_amount'] = safe_find_value(By.ID, 'form.trail_profit_target_percent_amount')
        settings['trail_profit_premium_trigger'] = safe_find_value(By.ID, 'form.trail_premium_value_profit_trigger')
        settings['trail_profit_premium_amount'] = safe_find_value(By.ID, 'form.trail_premium_value_profit_amount')
        settings['trailing_stop_sensitivity'] = safe_find_select(By.ID, 'form.trailing_stop_sensitivity')

        # --- Market Conditions Section ---
        settings['ma_crossover_toggle'] = safe_find_checkbox(By.ID, 'form.exit_ma_crossover_toggle')
        settings['ma_crossover_one'] = safe_find_select(By.ID, 'form.exit_ma_crossover_one')
        settings['ma_crossover_percent'] = safe_find_value(By.ID, 'form.exit_ma_crossover_percent')
        settings['ma_crossover_type'] = safe_find_select(By.ID, 'form.exit_ma_crossover_type')
        settings['ma_crossover_two'] = safe_find_select(By.ID, 'form.exit_ma_crossover_two')
        settings['exit_ma_value_toggle'] = safe_find_checkbox(By.ID, 'form.exit_ma_value_toggle')
        settings['ma_value_percent'] = safe_find_value(By.ID, 'form.exit_ma_value_percent')
        settings['ma_value_type'] = safe_find_select(By.ID, 'form.exit_ma_value_type')
        settings['ma_value_ma'] = safe_find_select(By.ID, 'form.exit_ma_value_ma')

        # --- Miscellaneous Section ---
        settings['exit_speed'] = safe_find_select(By.ID, 'form.exit_speed')
        settings['close_short_strike_only'] = safe_find_checkbox(By.ID, 'form.close_short_strike_only')

        return settings

    @_require_enabled
    def update_exit_settings(self, bot_num: str, exit_settings_dict: dict) -> dict:
        """
        Update WT Bot Exit Settings for the specified bot number using a dictionary of values.
        Updates all exit fields present in the UI, including checkboxes, selects, text fields, and repeaters.
        :param bot_num: WhisperTrades bot Number
        :param exit_settings_dict: Dictionary with the same format as returned by get_exit_settings
        :return: Dictionary of changed settings
        :rtype: dict
        """

        settings_keys = [
            # --- Profit (Standing) Section ---
            'profit_target_percent',
            'profit_premium_value',
            # --- Monitored Stops Section ---
            'stop_loss_percent',
            'loss_premium_value',
            'itm_percent_stop',
            'otm_percent_stop',
            'delta_stop',
            'monitored_stop_sensitivity',
            # --- Trailing Stops Section ---
            'trail_profit_percent_trigger',
            'trail_profit_percent_amount',
            'trail_profit_premium_trigger',
            'trail_profit_premium_amount',
            'trailing_stop_sensitivity',
            # --- Market Conditions Section ---
            'ma_crossover_toggle',
            'ma_crossover_one',
            'ma_crossover_percent',
            'ma_crossover_type',
            'ma_crossover_two',
            'exit_ma_value_toggle',
            'ma_value_percent',
            'ma_value_type',
            'ma_value_ma',
            # --- Miscellaneous Section ---
            'exit_speed',
            'close_short_strike_only',
        ]
        field_map = {
            # --- Profit (Standing) Section ---
            'profit_target_percent': ('form.profit_target_percent', 'text'),
            'profit_premium_value': ('form.premium_value_profit', 'text'),
            # --- Monitored Stops Section ---
            'stop_loss_percent': ('form.stop_loss_percent', 'text'),
            'loss_premium_value': ('form.premium_value_loss', 'text'),
            'itm_percent_stop': ('form.itm_percent', 'text'),
            'otm_percent_stop': ('form.otm_percent', 'text'),
            'delta_stop': ('form.delta_value', 'text'),
            'monitored_stop_sensitivity': ('form.monitored_stop_sensitivity', 'select'),
            # --- Trailing Stops Section ---
            'trail_profit_percent_trigger': ('form.trail_profit_target_percent_trigger', 'text'),
            'trail_profit_percent_amount': ('form.trail_profit_target_percent_amount', 'text'),
            'trail_profit_premium_trigger': ('form.trail_premium_value_profit_trigger', 'text'),
            'trail_profit_premium_amount': ('form.trail_premium_value_profit_amount', 'text'),
            'trailing_stop_sensitivity': ('form.trailing_stop_sensitivity', 'select'),
            # --- Market Conditions Section ---
            'ma_crossover_toggle': ('form.exit_ma_crossover_toggle', 'checkbox'),
            'ma_crossover_one': ('form.exit_ma_crossover_one', 'select'),
            'ma_crossover_percent': ('form.exit_ma_crossover_percent', 'text'),
            'ma_crossover_type': ('form.exit_ma_crossover_type', 'select'),
            'ma_crossover_two': ('form.exit_ma_crossover_two', 'select'),
            'exit_ma_value_toggle': ('form.exit_ma_value_toggle', 'checkbox'),
            'ma_value_percent': ('form.exit_ma_value_percent', 'text'),
            'ma_value_type': ('form.exit_ma_value_type', 'select'),
            'ma_value_ma': ('form.exit_ma_value_ma', 'select'),
            # --- Miscellaneous Section ---
            'exit_speed': ('form.exit_speed', 'select'),
            'close_short_strike_only': ('form.close_short_strike_only', 'checkbox'),
        }

        wt_exit_edit_url = f'https://whispertrades.com/bots/{bot_num}/exit/edit'
        self.__get_url_and_wait(wt_exit_edit_url)
        save_btn = self.webdriver.find_element(by=By.CLASS_NAME, value='bg-green-600')

        def safe_update_field(field_id, value, input_type='text'):
            try:
                el = WebDriverWait(self.webdriver, 2).until(EC.presence_of_element_located((By.ID, field_id)))
                if input_type == 'checkbox':
                    if bool(el.is_selected()) != bool(value):
                        el.click()
                elif input_type == 'select':
                    sel = Select(el)
                    try:
                        sel.select_by_value(str(value))
                    except Exception:
                        for option in sel.options:
                            if option.text.strip() == str(value).strip():
                                option.click()
                                break
                else:
                    # Use the known working __update_field for text fields, which now uses BACKSPACE
                    if value is None:
                        self.__update_field(el, "")
                    else:
                        self.__update_field(el, str(value))
                time.sleep(0.1)
            except Exception:
                pass


        # Update all fields in order, skip if value is None and field does not exist
        for k in settings_keys:
            v = exit_settings_dict.get(k, None)
            if v is None:
                # Only skip if field does not exist
                try:
                    self.webdriver.find_element(By.ID, field_map[k][0])
                except Exception:
                    continue
            safe_update_field(field_map[k][0], v, field_map[k][1])

        # Handle variables (repeater)
        if 'variables' in exit_settings_dict and isinstance(exit_settings_dict['variables'], list):
            for idx, var in enumerate(exit_settings_dict['variables']):
                prefix = f'form.exitVariables.{idx}.'
                if isinstance(var, dict):
                    if 'bot_variable_id' in var and var['bot_variable_id'] is not None:
                        try:
                            el = WebDriverWait(self.webdriver, 2).until(EC.presence_of_element_located((By.ID, prefix + 'bot_variable_id')))
                            sel = Select(el)
                            sel.select_by_value(str(var['bot_variable_id']))
                        except Exception as e:
                            if self.verbose:
                                print(f"[DEBUG] Exception updating exit variable bot_variable_id: {e}")
                    if 'condition' in var and var['condition'] is not None:
                        try:
                            el = WebDriverWait(self.webdriver, 2).until(EC.presence_of_element_located((By.ID, prefix + 'condition')))
                            sel = Select(el)
                            sel.select_by_value(str(var['condition']))
                        except Exception as e:
                            if self.verbose:
                                print(f"[DEBUG] Exception updating exit variable condition: {e}")
                    if 'value' in var and var['value'] is not None:
                        try:
                            el = WebDriverWait(self.webdriver, 2).until(EC.presence_of_element_located((By.ID, prefix + 'value')))
                            el.clear()
                            el.send_keys(str(var['value']))
                        except Exception as e:
                            if self.verbose:
                                print(f"[DEBUG] Exception updating exit variable value: {e}")

        save_btn.click()
        time.sleep(1)
        new_settings = self.get_exit_settings(bot_num)
        if self.verbose:
            print(f'Updated EXIT settings for bot: {bot_num}')
        if exit_settings_dict == new_settings:
            warnings.warn(f'Exit settings are unchanged for bot: {bot_num}')
        return new_settings

    @_require_enabled
    def enable_by_bot_num(self, bot_num: str) -> bool:
        """Enable bot by bot num

        :param bot_num: WhisperTrades bot Number
        :return: Bool value of successful url load
        :rtype: bool
        """
        url = f"https://whispertrades.com/bots/{bot_num}/enable"
        if self.verbose:
            print(f'Enabling bot: {bot_num}')
        return self.__get_url_and_wait(url)

    @_require_enabled
    def force_disable_by_bot_num(self, bot_num: str) -> bool:
        """Disable bot by bot num

        :param bot_num: WhisperTrades bot Number
        :return: Bool value of successful url load
        :rtype: bool
        """

        url = f"https://whispertrades.com/bots/{bot_num}/force_disable?redirectUrl=https%253A%252F%252Fwhispertrades.com%252Fbots"
        if self.verbose:
            print(f'Disabling bot: {bot_num}')
        return self.__get_url_and_wait(url)

    @_require_enabled
    def disable_on_close_by_bot_num(self, bot_num: str) -> bool:
        """Soft disable bot by bot num

        :param bot_num: WhisperTrades bot Number
        :return: Bool value of successful url load
        :rtype: bool
        """
        url = f"https://whispertrades.com/bots/{bot_num}/soft_disable?redirectUrl=https%253A%252F%252Fwhispertrades.com%252Fbots"
        if self.verbose:
            print(f'Disabling on close bot: {bot_num}')
        return self.__get_url_and_wait(url)

    @_require_enabled
    def enabled_to_soft_disabled_by_list(self, bot_num_lst: list[str]) -> None:
        """
        Bot status change: 'Enabled' to 'Disabled on Close'
        :param bot_num_lst: List of WhisperTrades bot Numbers
        :type url: list
        """
        for b in bot_num_lst:
            if not self.disable_on_close_by_bot_num(b):
                warnings.warn(f'UNABLE TO "DISABLE ON CLOSE" BOT_NUM {b}')
        return

    @_require_enabled
    def enabled_to_forced_disabled_by_list(self, bot_num_lst: list[str]) -> None:
        """
        Bot status change: 'Enabled' to 'Disabled'

        :param bot_num_lst: List of WhisperTrades bot Numbers
        :type url: list
        """
        for b in bot_num_lst:
            if not self.force_disable_by_bot_num(b):
                warnings.warn(f'UNABLE TO "DISABLE" BOT_NUM {b}')
        return

    @_require_enabled
    def disabled_to_enabled_by_list(self, bot_num_lst: list[str]) -> None:
        """
        Bot status change: 'Disabled' to 'Enabled'

        :param bot_num_lst: List of WhisperTrades bot Numbers
        :type url: list
        """
        for b in bot_num_lst:
            if not self.enable_by_bot_num(b):
                warnings.warn(f'UNABLE TO "ENABLE" BOT_NUM {b}')
        return

    @_require_enabled
    def force_enabled_by_list(self, bot_num_lst: list[str]) -> None:
        """
        Bot status change: to 'Enabled'

        :param bot_num_lst: List of WhisperTrades bot Numbers
        :type url: list
        """
        for b in bot_num_lst:
            if not self.enable_by_bot_num(b):
                warnings.warn(f'UNABLE TO "ENABLE" BOT_NUM {b}')
        return

    @_require_enabled
    def force_disable_by_list(self, bot_num_lst: list[str]) -> None:
        """
        Bot status change: Any state to 'Disabled'

        :param bot_num_lst: List of WhisperTrades bot Numbers
        :type url: list
        """
        for b in bot_num_lst:
            if not self.force_disable_by_bot_num(b):
                warnings.warn(f'UNABLE TO "DISABLE" BOT_NUM {b}')
        return
    
    @_require_enabled
    def enter_new_position_by_bot_num(self, bot_num: str) -> bool:
        """Enter new position by bot num

        :param bot_num: WhisperTrades bot Number
        :return: Bool value of successful url load
        :rtype: bool
        """
        url = f"https://whispertrades.com/bots/{bot_num}/enter_position"
        if self.verbose:
            print(f'Entering new position on bot: {bot_num}')
        return self.__get_url_and_wait(url)




########################################################################################################################
########################################################################################################################
###   Throttle Object for WhisperTrades.com API                                                                      ###
###                                                                                                                  ###
###   Authored by Paul Nobrega   Contact: Paul@PaulNobrega.net                                                       ###
###   Python Version 3.10                                                                                            ###
########################################################################################################################
########################################################################################################################
import time

class WhisperTradesThrottle(object):
    """
    API rate limiter for WhisperTrades.com.

    Attributes:
        delay_sec (int): Delay in seconds between requests.
        is_on (bool): Whether throttling is enabled.
    """
    def __init__(self) -> None:
        self.delay_sec: int = 2
        self.is_on: bool = True
   
    def __call__(self, fxn):
        return self.__run(fxn)

    def __run(self, fxn):
        if self.is_on:
            time.sleep(self.delay_sec)
        return fxn()
    
    def enable(self):
        """
        Toggle throttle to enabled
        """
        if self.is_on == False:
            self.is_on = True
        return 
    
    def disable(self):
        """
        Toggle throttle to disabled
        """
        if self.is_on == True:
            self.is_on = False
        return 
    
    def set_delay_sec(self, sec_delay: int=2):
        """
        Set seconds to delay before each API request

        :sec_delay: seconds to delay before each API request.  Default is 2 seconds to match WhisperTrades API rate limit
        :type sec_delay: int
        """
        self.delay_sec = sec_delay
        return




########################################################################################################################
# WhisperDriver: Python Wrapper for WhisperTrades.com API
#
# Author: Paul Nobrega   Contact: Paul@PaulNobrega.net
# Python Version: 3.10+
#
# This module provides a high-level Python interface to the WhisperTrades.com API, including
# bots, variables, reports, scheduling, and browser automation via Selenium.
#
# Usage Example:
#   from WhisperDriver import ApiWrapper
#   api = ApiWrapper(token="YOUR_API_TOKEN")
#   api.update_all_bots_list()
#   print(api.bot_number_list)
#
# Released under the MIT License.
########################################################################################################################

from . import Obj
import time
from typing import List, Optional

class ApiWrapper:
    """
    Main interface for interacting with the WhisperTrades.com API.

    Attributes:
        throttle: API rate limiter.
        endpts: Endpoint handler for all API endpoints.
        scheduler: Scheduler for timed tasks.
        bots: Bot management interface.
        variables: Variable management interface.
        via_selenium: Selenium-based UI automation interface.
        bot_number_list: List of all bot numbers.
        report_number_list: List of all report numbers.
        variable_number_list: List of all variable numbers.
    """
    def __init__(self, token: str):
        """
        Initialize the API wrapper with a WhisperTrades API token.

        Args:
            token (str): WhisperTrades.com API token.
        """
        self.throttle = Obj.WhisperTradesThrottle()
        self.throttle.disable()
        self.endpts = Obj.WhisperTradesEndpoints(token, self.throttle)
        self.scheduler = Obj.WhisperTradesScheduler(self.endpts)
        self.bots = Obj.WhisperTradesBots(self.scheduler)
        self.variables = Obj.WhisperTradesVariables(self.endpts)
        self.via_selenium = Obj.SeleniumDriver(self.endpts)
        self.bot_number_list: List[str] = []
        self.report_number_list: List[str] = []
        self.variable_number_list: List[str] = []
        self.__populate()
        self.throttle.enable()

    def __del__(self):
        """
        Ensure the scheduler thread is stopped before object destruction.
        """
        while self.scheduler.scheduler_is_on:
            time.sleep(1)
        return

    def __populate(self) -> None:
        """
        Populate all bot, report, and variable lists from the API.
        """
        self.update_all_bots_list()
        self.update_all_reports_list()
        self.update_all_variables_list()
        return

    def update_all_bots_list(self) -> None:
        """
        Update the list of bot numbers via the WhisperTrades API.
        """
        self.bots.update_all_bots()
        self.bot_number_list = [i.number for i in self.bots.bots_list.all]
        return

    def update_all_reports_list(self) -> None:
        """
        Update the list of report numbers via the WhisperTrades API.
        """
        self.report_number_list = [i['number'] for i in self.endpts.reports.get_all_bot_reports()]
        return

    def update_all_variables_list(self) -> None:
        """
        Update the list of variable numbers via the WhisperTrades API.
        """
        self.variable_number_list = [i['number'] for i in self.endpts.variables.get_all_bot_variables()]
        return

    def start_scheduler(self) -> None:
        """
        Start the scheduler loop in a unique thread. Thread automatically started at instantiation.
        """
        self.scheduler.start()
        return

    def stop_scheduler(self) -> None:
        """
        Stop the scheduler loop thread.
        """
        self.scheduler.stop()
        return

    def stop_scheduler_at_time(self, time_str: Optional[str] = None, tz_str: str = 'America/New_York') -> None:
        """
        Stop the scheduler thread at a predefined time.

        Args:
            time_str (Optional[str]): String representation of military time (e.g., '22:30'). If 12-hr format, PM or AM must be included.
            tz_str (str): Human-readable timezone. Default is 'America/New_York'.
        """
        self.scheduler.stop_scheduler_at_time(time_str, tz_str)
        return
        self.variable_number_list = [i['number'] for i in self.endpts.variables.get_all_bot_variables()]
    
    def start_scheduler(self):
        """
        Start Scheduler loop in unique thread.  Thread automatically started at instantiation
        """
        self.scheduler.start()
        return 
    
    def stop_scheduler(self):
        """
        Stop Scheduler loop thread.
        """
        self.scheduler.stop()
        return
    
    def stop_scheduler_at_time(self, time_str: str=None, tz_str: str='America/New_York'):
        """
        Stop Scheduler thread at predefined time.
        
        :param time_str: string representation of military time (example: '22:30'). If 12-hr format, PM or AM must be included in string.
        :type time_str: String
        :param tz_str: human readable TimeZone. Default is 'America/New_York'
        :type tz_str: String
        """
        self.scheduler.stop_scheduler_at_time(time_str, tz_str)
        return




########################################################################################################################
########################################################################################################################
###   Endpoint Objects for WhisperTrades.com API                                                                     ###
###                                                                                                                  ###
###   Authored by Paul Nobrgea   Contact: Paul@PaulNobrega.net                                                       ###
###   Python Version 3.10                                                                                            ###
########################################################################################################################
########################################################################################################################
import requests
from urllib.parse import urljoin
import posixpath
import json
import warnings



class WhisperTradesEndpoints(object):
    """
    Endpoint handler for WhisperTrades.com API.

    Args:
        token (str): API token for authentication.
        throttle (object): Throttle object for rate limiting.
    """
    def __init__(self, token: str, throttle: object) -> None:
        self.throttle: object = throttle
        self.config = self.__config(token)
        self.bots = self.__bots(self.config, self.throttle)
        self.reports = self.__reports(self.config, self.throttle)
        self.variables = self.__variables(self.config, self.throttle)
        self.brokers = self.__brokers(self.config, self.throttle)

    @staticmethod
    def format_response(response):
        """
        Format API responses to JSON and apply rate limit throttle
        """
        txt_to_json = json.loads(response.text)
        if not response.ok:
            msg = txt_to_json['message'] if 'message' in txt_to_json else ''
            warnings.warn(f"Status code: {response.status_code} received with reason: {response.reason} at url: {response.url}\n{msg}")
            if response.reason.lower() == 'unauthorized':
                raise Exception(f'Invalid API token!')
        return txt_to_json['data'] if 'data' in txt_to_json else txt_to_json

    class __config(object):
        def __init__(self, token):
            self.TOKEN = token
            self.SERVER = r'https://api.whispertrades.com/v1/'
            self.HEADERS = {'Accept': 'application/json', 'Content-Type': 'application/json', 'authorization': f'Bearer {self.TOKEN}'}
    
    class __bots(object):
    
        def __init__(self, config, throttle):
            self.endpt = r'bots/'
            self.config = config
            self._throttle = throttle
    
        def get_bot_orders(self, bot_number: str) -> json:
            """
            Get all orders for a specific bot by bot_number.

            :param bot_number: number of bot.
            :type bot_number: String

            :return: json data from response received from WhisperTrades API
            :type return: json
            """
            if not bot_number:
                raise ValueError("Bot number is required to get orders.")
            url_path = urljoin(self.config.SERVER, posixpath.join(self.endpt, bot_number, 'orders'))
            payload = {}
            response = self._throttle(lambda: requests.request("GET", url_path, headers=self.config.HEADERS, data=payload))
            return WhisperTradesEndpoints.format_response(response)

        def get_bot_positions(self, bot_number: str = '', position_number: str = '', status: str = '', from_date: str = '', to_date: str = '', page: str = '') -> json:
            """
            Get a paginated list of bot positions, or a single position if position_number is provided.

            :param bot_number: Bot number to filter positions (optional)
            :param position_number: Unique position number to retrieve a single position (optional)
            :param status: Filter by position status (OPEN or CLOSED, optional)
            :param from_date: Minimum entry date (YYYY-MM-DD, optional)
            :param to_date: Maximum entry date (YYYY-MM-DD, optional)
            :param page: Page number (optional)
            :return: json data from response received from WhisperTrades API
            """
            if position_number:
                url_path = urljoin(self.config.SERVER, posixpath.join('bots/positions', position_number))
                params = {}
            else:
                url_path = urljoin(self.config.SERVER, 'bots/positions')
                params = {}
                if bot_number:
                    params['bot'] = bot_number
                if status:
                    params['status'] = status
                if from_date:
                    params['from_date'] = from_date
                if to_date:
                    params['to_date'] = to_date
                if page:
                    params['page'] = page
            response = self._throttle(lambda: requests.request("GET", url_path, params=params, headers=self.config.HEADERS))
            return WhisperTradesEndpoints.format_response(response)

        def close_bot_position(self, position_number: str) -> json:
            """
            Close a specific bot position by position number.

            :param position_number: Position number to close
            :return: json data from response received from WhisperTrades API
            """
            if not position_number:
                raise ValueError("Position number is required to close a bot position.")
            url_path = urljoin(self.config.SERVER, posixpath.join('bots/positions', position_number, 'close'))
            response = self._throttle(lambda: requests.request("PUT", url_path, headers=self.config.HEADERS))
            return WhisperTradesEndpoints.format_response(response)

        def get_bot(self, bot_number: str='', status_filter: list=['Enabled', 'Disabled', 'Disable on Close'], include_details: bool=True) -> json:
            """
            Query WhisperTrades.com for bot information at endpoint: 'bots/' 

            :param bot_number: number of bot. Default = '' to return all bot information
            :type bot_number: String
            :param status_filter: Empty list or list containing any combination of: 'Enabled', ' Disabled', 'Disabled on Close'. Default = [] to apply no filter
            :type password: List
            :param include_details: Include all bot settings. Default = True
            :type include_details: Boolean

            :return: json data from response received from WhisperTrades API
            :type return: json
            """
            url_path = urljoin(self.config.SERVER, posixpath.join(self.endpt, bot_number))
            payload={}
            params = {'include_details':include_details, 'statuses[]': status_filter}
            response = self._throttle(lambda: requests.request("GET", url_path, params=params, headers=self.config.HEADERS, data=payload))
            return WhisperTradesEndpoints.format_response(response)
        
        def get_all_bots(self) -> json:
            """
            Query WhisperTrades.com for ALL bot information at endpoint: 'bots/' 

            :return: json data from response recieved from WhisperTrades API
            :type return: json
            """
            return self.get_bot(bot_number='', status_filter=[])
        
        def enable_bot(self, bot_number:str) -> json:
            """
            Enable WhisperTrades.com bot by bot_number 

            :param bot_number: number of bot. Default = '' to return all bot information
            :type bot_number: String

            :return: json data from response recieved from WhisperTrades API
            :type return: json
            """
            url_path = urljoin(self.config.SERVER, posixpath.join(self.endpt, bot_number,'enable'))
            payload={}
            response = self._throttle(lambda: requests.request("PUT", url_path, headers=self.config.HEADERS, data=payload))
            return WhisperTradesEndpoints.format_response(response)

        def disable_bot(self, bot_number:str) -> json:
            """
            Disable WhisperTrades.com bot by bot_number.  If bot has open positions, status will be set to 'Disable on Close'

            :param bot_number: number of bot. Default = '' to return all bot information
            :type bot_number: String

            :return: json data from response recieved from WhisperTrades API
            :type return: json
            """
            url_path = urljoin(self.config.SERVER, posixpath.join(self.endpt, bot_number,'disable'))
            payload={}
            response = self._throttle(lambda: requests.request("PUT", url_path, headers=self.config.HEADERS, data=payload))
            return WhisperTradesEndpoints.format_response(response) 

        def open_position(self, bot_number: str) -> json:
            """
            Open a position for the specified bot.

            :param bot_number: number of bot.
            :type bot_number: String

            :return: json data from response received from WhisperTrades API
            :type return: json
            """
            url_path = urljoin(self.config.SERVER, posixpath.join(self.endpt, bot_number, 'open_position'))
            payload = {}
            response = self._throttle(lambda: requests.request("PUT", url_path, headers=self.config.HEADERS, data=payload))
            return WhisperTradesEndpoints.format_response(response)

        def close_position(self, bot_number: str) -> json:
            """
            Close a position for the specified bot.

            :param bot_number: number of bot.
            :type bot_number: String

            :return: json data from response received from WhisperTrades API
            :type return: json
            """
            url_path = urljoin(self.config.SERVER, posixpath.join(self.endpt, bot_number, 'close_position'))
            payload = {}
            response = self._throttle(lambda: requests.request("PUT", url_path, headers=self.config.HEADERS, data=payload))
            return WhisperTradesEndpoints.format_response(response)

    class __reports(object):
        
        def __init__(self, config, throttle):
            self.endpt = r'bots/reports/'
            self.config = config
            self._throttle = throttle
        
        def get_bot_report(self, report_number: str='') -> json:
            """
            Query WhisperTrades.com for report by report_number.

            :param report_number: number of report. Default = '' to return all report information
            :type report_number: String

            :return: json data from response recieved from WhisperTrades API
            :type return: json
            """
            url_path = urljoin(self.config.SERVER, posixpath.join(self.endpt, report_number))
            payload={}
            response = self._throttle(lambda: requests.request("GET", url_path, headers=self.config.HEADERS, data=payload))
            return WhisperTradesEndpoints.format_response(response)
        
        def get_all_bot_reports(self) -> json:
            """
            Query WhisperTrades.com for ALL reports.

            :return: json data from response recieved from WhisperTrades API
            :type return: json
            """
            return self.get_bot_report('')
        
        def update_bot_report(self, report_number: str, new_name:str = '', new_start_date:str='', new_end_date:str='', run_until_latest_date:bool=False) -> json:
            """
            Query WhisperTrades.com for report by report_number.
            
            Required
            :param report_number: number of report.
            :type report_number: String

            Optional
            :param new_name: New name to rename report as
            :type new_name: String
            :param new_start_date: New start date for the report. Expected format is 'YYYY-MM-DD'
            :type new_start_date: String
            :param new_end_date: New end date for the report. Expected format is 'YYYY-MM-DD'
            :type new_end_date: String
            :param run_until_latest_date: Run report until current date if True. Default = False
            :type run_until_latest_date: Boolean

            :return: json data from response recieved from WhisperTrades API
            :type return: json
            """
            url_path = urljoin(self.config.SERVER, posixpath.join(self.endpt, report_number))
            payload = {
                "name": new_name,
                "start_date": new_start_date,
                "end_date": new_end_date,
                "run_until_latest_date": run_until_latest_date
                }
            
            for key in payload:
                if payload[key] == '':
                    del payload[key]

            response = self._throttle(lambda: requests.request("PUT", url_path, headers=self.config.HEADERS, data=json.dumps(payload)))
            return WhisperTradesEndpoints.format_response(response)
        
        def run_bot_report(self, report_number:str=''):
            """
            Run report at WhisperTrades.com by report_number.
            
            Required
            :param report_number: number of report.
            :type report_number: String

            :return: json data from response recieved from WhisperTrades API
            :type return: json
            """

            if report_number == '':
                raise ValueError(f"Insufficient information supplied to run report! (REQUIRED) report_number: {report_number}")
            url_path = urljoin(self.config.SERVER, posixpath.join(self.endpt, report_number,'run'))
            payload={}
            response = self._throttle(lambda: requests.request("PUT", url_path, headers=self.config.HEADERS, data=payload))
            return WhisperTradesEndpoints.format_response(response)
    
    class __variables(object):
        
        def __init__(self, config, throttle):
            self.endpt = r'bots/variables/'
            self.config = config
            self._throttle = throttle
        
        def get_bot_variables(self, variable_number:str='') -> json:
            """
            Query WhisperTrades.com for variable by variable_number.

            :param variable_number: number of variable. Default = '' to return all variable information
            :type variable_number: String

            :return: json data from response recieved from WhisperTrades API
            :type return: json
            """
            url_path = urljoin(self.config.SERVER, posixpath.join(self.endpt, variable_number))
            payload={}
            response = self._throttle(lambda: requests.request("GET", url_path, headers=self.config.HEADERS, data=payload))
            return WhisperTradesEndpoints.format_response(response)
        
        def get_all_bot_variables(self) -> json:
            """
            Query WhisperTrades.com for ALL variables.

            :return: json data from response recieved from WhisperTrades API
            :type return: json
            """
            return self.get_bot_variables('')
        
        def set_bot_variables(self, variable_number:str='', variable_name:str='', new_value:str='') -> json:
            """
            Set WhisperTrades.com variable by name to new given value.

            REQUIRED
            :param variable_number: number of variable.
            :type variable_number: String
            :param variable_name: name of variable.
            :type variable_name: String
            :param new_value: new 'free text type' value to associate with variable name.
            :type new_value: String

            :return: json data from response recieved from WhisperTrades API
            :type return: json
            """
            if variable_number == '' or variable_name == '' or new_value == '':
                raise ValueError(f"Insufficient information supplied to set bot variable! (REQUIRED) variable_number: {variable_number}, (REQUIRED) variable_name: {variable_name}, (REQUIRED) new_value: {new_value}")
            url_path = urljoin(self.config.SERVER, posixpath.join(self.endpt, variable_number))
            payload = json.dumps({"name": variable_name, "value": new_value})
            response = self._throttle(lambda: requests.request("PUT", url_path, headers=self.config.HEADERS, data=payload))
            return WhisperTradesEndpoints.format_response(response)
    
    class __brokers(object):
        
        def __init__(self, config, throttle):
            self.endpt = r'broker_connections/'
            self.config = config
            self._throttle = throttle
        
        def get_broker_connections(self, number: str = '') -> json:
            """
            Get a single broker connection or a list of all broker connections.

            :param number: Unique identifier to retrieve a single broker connection. If omitted, all broker connections will be returned.
            :type number: String

            :return: json data from response received from WhisperTrades API
            :type return: json
            """
            url_path = urljoin(self.config.SERVER, posixpath.join(self.endpt, number))
            payload = {}
            response = self._throttle(lambda: requests.request("GET", url_path, headers=self.config.HEADERS, data=payload))
            data = WhisperTradesEndpoints.format_response(response)
            # Remove any broker connections with 'broker': 'TD Ameritrade' as broker is no longer valid
            if isinstance(data, list):
                data = [d for d in data if d.get('broker', '').lower() != 'td ameritrade']
            elif isinstance(data, dict) and data.get('broker', '').lower() == 'td ameritrade':
                return None
            return data

        def get_all_broker_connections(self) -> json:
            """
            Get a list of all broker connections.

            :return: json data from response received from WhisperTrades API
            :type return: json
            """
            return self.get_broker_connections('')

        def rebalance_broker_collateral(self, number: str) -> json:
            """
            Rebalance your collateral position for a given broker connection.
            This requires that the collateral be configured and enabled at Whispertrades.
            If your current collateral balance is within the minimum and maximum target amounts, a transaction will not happen.

            :param number: Unique identifier for the broker connection to rebalance
            :type number: String

            :return: json data from response received from WhisperTrades API
            :type return: json
            """
            if not number:
                raise ValueError("Broker connection number is required for collateral rebalance.")
            url_path = urljoin(self.config.SERVER, posixpath.join(self.endpt, number, 'collateral', 'rebalance'))
            payload = {}
            response = self._throttle(lambda: requests.request("PUT", url_path, headers=self.config.HEADERS, data=payload))
            return WhisperTradesEndpoints.format_response(response)


import os

def get_hour_minute_ampm_format():
    """
    Returns the strftime/strptime format string for hour:minute AM/PM (cross-platform, always valid for strptime).
    Use '%I:%M %p' for parsing times with strptime.
    """
    return '%I:%M %p'



########################################################################################################################
########################################################################################################################
###   Scheduler Object for WhisperTrades.com API                                                                     ###
###                                                                                                                  ###
###   Authored by Paul Nobrega   Contact: Paul@PaulNobrega.net                                                       ###
###   Python Version 3.10                                                                                            ###
########################################################################################################################
########################################################################################################################
import schedule
from datetime import datetime
from WhisperDriver.utils.time import get_hour_minute_ampm_format
import time
from threading import Thread

class WhisperTradesScheduler(object):
    """
    Scheduler for WhisperTrades.com API bot actions.

    Args:
        endpts (object): Endpoints object for API access.
    """
    def __init__(self, endpts: object) -> None:
        self._endpts: object = endpts
        self.scheduler_is_on: bool = False
        self.__scheduler = None
   
    def __del__(self):
        self.stop()
        return

    def __schedule_loop(self):
        while self.scheduler_is_on:
            schedule.run_pending()
            time.sleep(1)
        return
    
    def start(self):
        """
        Start Scheduler loop in unique thread.  Thread automatically started at instantiation
        """
        if self.scheduler_is_on == False:
            self.scheduler_is_on = True
            self.__scheduler = None
            self.__scheduler = Thread(target=self.__schedule_loop)
            self.__scheduler.start()
        return 
    
    def stop(self):
        """
        Stop Scheduler loop thread.
        """
        if self.scheduler_is_on == True:
            schedule.clear()
            self.scheduler_is_on = False
            time.sleep(1)
        return
    
    def stop_scheduler_at_time(self, time_str: str=None, tz_str: str='America/New_York'):
        """
        Stop Scheduler thread at predefined time.
        
        :param time_str: string representation of military time (example: '22:30'). If 12-hr format, PM or AM must be included in string.
        :type time_str: String
        :param tz_str: human readable TimeZone. Default is 'America/New_York'
        :type tz_str: String
        """
        if not time_str or not isinstance(time_str, str):
            raise ValueError('Time input string is required!')
        if 'pm' in time_str.lower() or 'am' in time_str.lower():
            time_str = datetime.strptime(time_str, get_hour_minute_ampm_format()).strftime('%H:%M')
        self.start()
        schedule.every().day.at(time_str, tz_str).do(self.stop)
        return

    def add_task(self, time_str: str=None, tz_str: str='America/New_York', fxn=None):
        """
        Add Task to scheduler.
        
        :param time_str: string representation of military time (example: '22:30'). If 12-hr format, PM or AM must be included in string.
        :type time_str: String
        :param tz_str: human readable TimeZone. Default is 'America/New_York'
        :type tz_str: String
        :param fxn: function to execute
        :type tz_str: python function
        """
        if not time_str or not isinstance(time_str, str):
            raise ValueError('Time input string is required!')
        if 'pm' in time_str.lower() or 'am' in time_str.lower():
            time_str = datetime.strptime(time_str, get_hour_minute_ampm_format()).strftime('%H:%M')
        schedule.every().day.at(time_str, tz_str).do(fxn)
        return




########################################################################################################################
########################################################################################################################
###   Bot Objects for WhisperTrades.com API                                                                          ###
###                                                                                                                  ###
###   Authored by Paul Nobrega   Contact: Paul@PaulNobrega.net                                                       ###
###   Python Version 3.10                                                                                            ###
########################################################################################################################
########################################################################################################################
import json
import warnings
from datetime import datetime


class WhisperTradesBots(object):
    """
    Bot handler for WhisperTrades.com API.

    Args:
        scheduler (object): Scheduler object for bot scheduling and endpoint access.
    """
    def __init__(self, scheduler: object) -> None:
        self._scheduler: object = scheduler
        self._endpts = self._scheduler._endpts
        self.bots_list = self.__bot_list(self._scheduler)

    def __call__(self, bot_number):
        """
        Return the bot object for the provided bot number.
        Usage: WD.bots('BOT_NUMBER')
        """
        for bot in self.bots_list.all:
            if hasattr(bot, 'number') and bot.number == bot_number:
                return bot
        raise KeyError(f"Bot number '{bot_number}' not found.")
       
    def get_all_bot_variables(self) -> json:
        """
        Query WhisperTrades.com for all bot variables and associate data with related bot object
       
        :return: json data from response recieved from WhisperTrades API
        :type return: json
        """
        all_variables = self._endpts.variables.get_all_bot_variables()
        for var in all_variables:
            for bot in self.bots_list.all:
                for i, bot_var in enumerate(bot.variables):
                    if bot_var['number'] == var['number']:
                        bot.variables[i] = var
        return all_variables
    
    def update_all_bots(self):
        """
        Update bots_list with data retrieved from WhisperTrades.com API
        """
        self.bots_list.all = []
        _ = [self.bots_list.add_bot_to_list(bot) for bot in self._endpts.bots.get_all_bots()]
        self.bots_list.update_positions_all_bots()
        return

    class __bot_list(object):

        def __init__(self, scheduler):
            self.all = []
            self._scheduler = scheduler
            self._endpts = self._scheduler._endpts
        
        def __call__(self):
            """
            Return list of all bot numbers
            """
            return self.all
        
        def all(self) -> list:
            """
            Return list of all bot numbers
            """
            return self.all
        
        def is_enabled(self) -> list:
            """
            Return list of all bot numbers that have status = 'enabled'
            """
            return [bot for bot in self.all if bot.status.lower() == 'enabled']
        
        def is_disabled(self) -> list:
            """
            Return list of all bot numbers that have status = 'disabled'
            """
            return [bot for bot in self.all if bot.status.lower() == 'disabled']
        
        def is_disabled_on_close(self) -> list:
            """
            Return list of all bot numbers that have status = 'disabled on close'
            """
            return [bot for bot in self.all if bot.status.lower() == 'disabled on close']
        
        def add_bot_to_list(self, bot_dict:dict={}):
            """
            Add dictionary representation of a WT bot to bot_list.all

            Note: if bot_number exists in bot_list.all, it is removed and replaced with the new information
            """
            if bot_dict=={}:
                warnings.warn(f'bot_dict is empty!')
                return
            bot_json = json.loads(json.dumps(bot_dict))
            self.remove_bot_from_list(bot_json['number'])
            self.all.append(self.bot_obj(bot_json, self._scheduler))
            return
        
        def remove_bot_from_list(self, bot_number:str):
            """
            Removes bot from bots.all list by given bot number
            """
            for i in range(len(self.all)):
                if self.all[i].number == bot_number:
                    del self.all[i]
                    return
            return
        
        def update_positions_all_bots(self):
            """
            Fetch all positions for all bots in one API call and update each bot's .positions attribute accordingly.
            """

            try:
                all_positions = self._endpts.bots.get_bot_positions()
                if isinstance(all_positions, dict) and 'data' in all_positions:
                    all_positions = all_positions['data']
                # Build a mapping from bot number to list of positions
                bot_positions_map = {}
                for pos in all_positions:
                    bot_info = pos.get('bot') if isinstance(pos, dict) else None
                    bot_number = bot_info.get('number') if bot_info else None
                    if bot_number:
                        bot_positions_map.setdefault(bot_number, []).append(pos)
                # Assign positions to each bot in the list
                for bot in self.all:
                    bot.positions = bot_positions_map.get(bot.number, [])
            except Exception as e:
                for bot in self.all:
                    bot.positions = []
                print(f"Failed to update positions for all bots: {e}")


        class bot_obj(object):
           
            def __init__(self, bot_dict, scheduler):
                self.number = ''
                self.name = ''
                self.broker_connection = {}
                self.is_paper= False
                self.status = ''
                self.can_enable = True
                self.can_disable = True
                self.symbol = ''
                self.type = ''
                self.notes = ''
                self.last_active_at = ''
                self.disabled_at = ''
                self.entry_condition = {}
                self.exit_condition = {}
                self.adjustments = []
                self.notifications = []
                self.variables = []
                self.positions = []  # List of positions for this bot
                self._scheduler = scheduler
                self._endpts = self._scheduler._endpts
                self.__bot_dict_to_attr(bot_dict)                
            
            def _meridian_time_to_military_time(self, time_str):
                from WhisperDriver.utils.time import get_hour_minute_ampm_format
                return datetime.strptime(time_str, get_hour_minute_ampm_format()).strftime('%H:%M')
            
            def enable(self):
                """
                Enable the bot immediately.
                """
                return self._change_status('enable', self.number, self._endpts, self._scheduler)()

            def disable(self):
                """
                Disable the bot immediately.
                """
                return self._change_status('disable', self.number, self._endpts, self._scheduler)()
            
            def enable_at_time(self, time_str, tz_str='America/New_York'):
                """
                Schedule bot status change.
                
                :param time_str: string representation of military time (example: '22:30'). If 12-hr format, PM or AM must be included in string.
                :type time_str: String
                :param tz_str: human readable TimeZone. Default is 'America/New_York'
                :type tz_str: String
                """
                if not time_str or not isinstance(time_str, str):
                    raise ValueError('Time input string is required!')
                if 'pm' in time_str.lower() or 'am' in time_str.lower():
                    time_str = self._meridian_time_to_military_time(time_str)
                if not self._scheduler.scheduler_is_on:
                    self._scheduler.start()
                self._scheduler.add_task(time_str, tz_str, self.enable)
                return

            def disable_at_time(self, time_str, tz_str='America/New_York'):
                """
                Schedule bot status change.
                
                :param time_str: string representation of military time (example: '22:30'). If 12-hr format, PM or AM must be included in string.
                :type time_str: String
                :param tz_str: human readable TimeZone. Default is 'America/New_York'
                :type tz_str: String
                """
                if not time_str or not isinstance(time_str, str):
                    raise ValueError('Time input string is required!')
                if 'pm' in time_str.lower() or 'am' in time_str.lower():
                    time_str = self._meridian_time_to_military_time(time_str)
                if not self._scheduler.scheduler_is_on:
                    self._scheduler.start()
                self._scheduler.add_task(time_str, tz_str, self.disable)
                return

            def get_positions(self, position_number: str = '', status: str = '', from_date: str = '', to_date: str = '', page: str = ''):
                """
                Get all positions for this bot, or a single position if position_number is provided.
                """
                return self._endpts.bots.get_bot_positions(
                    bot_number=self.number,
                    position_number=position_number,
                    status=status,
                    from_date=from_date,
                    to_date=to_date,
                    page=page
                )

            def close_position_by_number(self, position_number: str):
                """
                Close a specific position by position number for this bot.
                """
                return self._endpts.bots.close_bot_position(position_number)
            
            def get_orders(self):
                """
                Get all orders for this bot using the API endpoint.
                """
                return self._endpts.bots.get_bot_orders(self.number)
            def open_position(self):
                """
                Open a position for this bot using the API endpoint.
                """
                return self._endpts.bots.open_position(self.number)

            def close_position(self):
                """
                Close a position for this bot using the API endpoint.
                """
                return self._endpts.bots.close_position(self.number)
           
            def __str__(self):
                attrs = vars(self)
                test = [f'{item[0]}: {str(item[1])}' for item in attrs.items()]
                return "\n".join(test)
            
            def __repr__(self):
                return self.__str__()
            
            def __bot_dict_to_attr(self, bot_dict):
                for key in bot_dict: 
                    setattr(self, key, bot_dict[key])
            
            def _refresh_positions(self):
                """
                Query WhisperTrades.com for all positions for this bot and update self.positions,
                only including positions where the bot number matches this bot.
                """
                try:
                    positions_data = self._endpts.bots.get_bot_positions(bot_number=self.number)
                    # Extract positions list from API response
                    if isinstance(positions_data, dict) and 'data' in positions_data:
                        all_positions = positions_data['data']
                    else:
                        all_positions = positions_data
                    # Filter positions to only those with matching bot number
                    filtered = []
                    for pos in all_positions:
                        bot_info = pos.get('bot') if isinstance(pos, dict) else None
                        if bot_info and bot_info.get('number') == self.number:
                            filtered.append(pos)
                    self.positions = filtered
                except Exception as e:
                    self.positions = []
                    print(f"Failed to fetch positions for bot {self.number}: {e}")

            def update(self):
                """
                Query WhisperTrades.com for bot information and update object with new information, including positions
                """
                bot_dict = self._endpts.bots.get_bot(bot_number=self.number)
                self.__bot_dict_to_attr(json.loads(json.dumps(bot_dict)))
                self._refresh_positions()
                return
           
            def get_bot_variables(self):
                """
                Query WhisperTrades.com for variables associated with bot and update object with new information 
                """
                all_var = [v['number'] for v in self.variables]
                self.variables = []
                self.variables = [self._endpts.variables.get_bot_variables(v) for v in all_var]
                return self.variables
            
            class _change_status(object):

                def __init__(self, target_status, bot_number, endpts, scheduler):
                    self._endpts = endpts
                    self._scheduler = scheduler
                    self._target_status = target_status
                    self._bot_number = bot_number

                def __call__(self):
                    return self._toggle_status()
                
                def _toggle_status(self):
                    if self._target_status == 'enable':
                        print(f"Enabling bot: {self._bot_number}")
                        return self._endpts.bots.enable_bot(self._bot_number)
                    elif self._target_status == 'disable':
                        print(f"Disabling bot: {self._bot_number}")
                        return self._endpts.bots.disable_bot(self._bot_number)
                    return
                
                    



    


from setuptools import setup, find_packages

with open("requirements.txt", encoding="utf-8") as f:
    requirements = [line.strip() for line in f if line.strip() and not line.startswith("#")]

setup(
    name="WhisperDriver",
    version="0.1.8",
    description="Automation and API library for WhisperTrades.com",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    author="Paul Nobrega",
    author_email="Paul@PaulNobrega.net",
    url="https://github.com/PaulNobrega/WhisperDriver",
    packages=find_packages(),
    install_requires=requirements,
    python_requires=">=3.10",
    license="MIT",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    include_package_data=True,
)



########################################################################################################################
########################################################################################################################
###   Variable Object for WhisperTrades.com API                                                                      ###
###                                                                                                                  ###
###   Authored by Paul Nobrega   Contact: Paul@PaulNobrega.net                                                       ###
###   Python Version 3.10                                                                                            ###
########################################################################################################################
########################################################################################################################
import json
import warnings

class WhisperTradesVariables(object):
    """
    Handles bot/global variable API actions for WhisperTrades.com.

    Args:
        endpts (object): Endpoints object for API calls.
    """
    def __init__(self, endpts: object) -> None:
        self._endpts: object = endpts
        self.unassociated_variable_numbers: list = []
        self.variables_list = self.__variable_list(self._endpts)
        self.update_all_variables()
    
    def __call__(self, variable_number: str):
        """
        Return the variable object for the provided variable number.
        Usage: WD.variables('VARIABLE_NUMBER')
        """
        for vari in self.variables_list.all:
            if hasattr(vari, 'number') and vari.number == variable_number:
                return vari
        raise KeyError(f"Variable number '{variable_number}' not found.")

    def update_all_variables(self):
        """
        Update variables_list with data retrieved from WhisperTrades.com API
        """
        self.variables_list.all = []
        _ = [self.variables_list.add_variable_to_list(vari) for vari in self._endpts.variables.get_all_bot_variables()]
        self.unassociated_variable_numbers = [vari.number for vari in self.variables_list.all if vari.bot is None]
        return
    

    class __variable_list(object):
            def __init__(self, endpts):
                self.all = []
                self._endpts = endpts
            
            def all(self) -> list:
                """
                Return list of all bot numbers
                """
                return self.all
            
            def add_variable_to_list(self, variable_dict:dict={}):
                """
                Add dictionary representation of a WT variable to variable_list.all

                Note: if variable_number exists in variables_list.all, it is removed and replaced with the new information
                """
                if variable_dict=={}:
                    warnings.warn(f'variable_dict is empty!')
                    return
                vari_json = json.loads(json.dumps(variable_dict))
                self.remove_variable_from_list(vari_json['number'])
                self.all.append(self.vari_obj(vari_json, self._endpts))
                return
            
            def remove_variable_from_list(self, variable_number:str):
                """
                Removes variable from variables_list.all list by given variable number
                """
                for i in range(len(self.all)):
                    if self.all[i].number == variable_number:
                        del self.all[i]
                        return
                return

            class vari_obj(object):
                
                def __init__(self, vari_dict, endpts):
                    self.number = ''
                    self.name = ''
                    self.bot = ''
                    self.value= ''
                    self.free_text_value = ''
                    self.last_updated_at = ''
                    self.conditions = []
                    self._endpts = endpts
                    self.__vari_dict_to_attr(vari_dict)
            
                def __str__(self):
                    attrs = vars(self)
                    test = [f'{item[0]}: {str(item[1])}' for item in attrs.items()]
                    return "\n".join(test)
                
                def __repr__(self):
                    return self.__str__()
                
                def __vari_dict_to_attr(self, vari_dict):
                    for key in vari_dict: 
                        setattr(self, key, vari_dict[key])
                
                def update(self):
                    """
                    Query WhisperTrades.com for bot information and update object with new information 
                    """
                    vari_dict = self._endpts.variables.get_bot_variables(variable_number = self.number)
                    self.__vari_dict_to_attr(json.loads(json.dumps(vari_dict)))
                    return
                
                def set(self, new_value:str='') -> json:
                    """
                    Set 'unassociated' bot variables via WhisperTrades.com, update object with new information, return json

                    REQUIRED
                    :param new_value: new 'free text type' value to associate with variable name.
                    :type variable_number: String

                    :return: json data from response received from WhisperTrades API
                    :type return: json
                    """
                    vari_dict = self._endpts.variables.set_bot_variables(variable_number=self.number, variable_name=self.name, new_value=new_value)
                    self.__vari_dict_to_attr(json.loads(json.dumps(vari_dict)))
                    return vari_dict



        